import matplotlib.pyplot as plt
import numpy as np
import os

# This script runs the cosmosis test sampler for a range of input HOD parameters and plots the IA models

# These are the parameters that I want to change and the range that I want to vary them over
# I'm using the prior ranges here from Dvornik et al 2023 https://arxiv.org/pdf/2210.03110.pdf
hod_params={
'norm_c':[0.1,1.0],
'log_ml_0':[7.0,13.0],
'log_ml_1':[9.0,14.0],
'g1':[2.5,15.0],
'g2':[0.1,10.0],
'scatter':[0.0,2.0],
'norm_s':[0.1,1.0],
'pivot': [10.0,14.0],
'b0':[-5.0,5.0],
'alpha_s':[-5.0,5.0],
'b1':[-5.0,5.0],
'b2':[-5.0,5.0]
}

isample=10
gamma_hat_params={
'nmass':np.linspace(5,50,num=2+isample).astype(int),
'kmin':np.geomspace(0.0001,0.1,num=2+isample),
'kmax':np.geomspace(10,10000,num=2+isample),
'nk':np.linspace(10,50,num=2+isample).astype(int)
}

# A plotting module
datadir='output/HM_IA/'
def plot_cl_GIII_over_GG(ax,iz,jz,colour):
    # Read in the power spectra files
    shearcl='shear_cl_gg'
    ellvals = np.loadtxt(datadir+shearcl+'/ell.txt')
    GGvals = np.loadtxt(datadir+shearcl+'/bin_'+str(iz)+'_'+str(jz)+'.txt')
    #II and GI
    #IIvals = np.loadtxt(datadir+'shear_cl_ii/bin_'+str(iz)+'_'+str(jz)+'.txt')
    #GIvals = np.loadtxt(datadir+'shear_cl_gi/bin_'+str(iz)+'_'+str(jz)+'.txt')
    #ratio = (IIvals + GIvals + GGvals)/GGvals
    shearcl='shear_cl'  #This is GG+GI(ij) + GI(ji) +II
    clvals = np.loadtxt(datadir+shearcl+'/bin_'+str(iz)+'_'+str(jz)+'.txt')
    ratio = clvals/GGvals
    print(clvals[1],ratio[1])
    ax.plot(ellvals, ratio, color=colour,linestyle='solid')


def plot_cl_shear_over_default(shearcl,ax,iz,jz,colour):
    # Read in the power spectra files
    ellvals = np.loadtxt(datadir+shearcl+'/ell.txt')
    clvals = np.loadtxt(datadir+shearcl+'/bin_'+str(iz)+'_'+str(jz)+'.txt')
    default_vals = np.loadtxt('../'+datadir+shearcl+'/bin_'+str(iz)+'_'+str(jz)+'.txt')
    ratio = clvals/default_vals -1.0
    print(clvals[1],ratio[1])
    ax.plot(ellvals, ratio, color=colour,linestyle='solid')
    ax.yaxis.get_major_formatter().set_useOffset(False)


def plot_tomobins(ip,iv,shearcl,ax):
    plot_cl_shear_over_default(shearcl,ax[ip,0],1,1,colours[iv])  #tomo bin 1-1
    plot_cl_shear_over_default(shearcl,ax[ip,1],5,1,colours[iv])  #tomo bin 1-5
    plot_cl_shear_over_default(shearcl,ax[ip,2],5,5,colours[iv])  #tomo bin 5-5

def put_exponent_in_top_right_corner(ax):
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    ax.get_yaxis().get_offset_text().set_visible(False)
    ax_max = max(ax.get_yticks())
    exponent_axis = np.floor(np.log10(ax_max)).astype(int)
    ax.text(0.01,0.01,r'$\times$10$^{%i}$'%(exponent_axis),
             ha='left', va='bottom',transform = ax.transAxes,fontsize=12)

def tidy_my_plot(ax,ip):
    # The exponent ends up overlapping with the upper panel plot :(
    # So we're going to use this slightly messy hack to remove it
    for i in range(3):
        put_exponent_in_top_right_corner(ax[ip,i])
    #Put the title in the bottom left corner
    ax[ip,0].text(0.01,1.1,param,ha='left', va='top',transform = ax[ip,0].transAxes,fontsize=12)
    ax[0,0].set_xscale('log')
    ax[ip,0].set_ylabel(r'$\Delta C(\ell)/C(\ell)$',fontsize=12)



#How many samples do you want to take in addition to testing the minimum and maximum values?
#Set up a nice range of colours to plot the samples in
colours = plt.cm.jet(np.linspace(0,1,isample+2))

#We're going to make separate figures for GG+GI+II,  GI and II
figall, axall = plt.subplots(4, 3,figsize=(6,8),sharex='all')#,gridspec_kw=dict(hspace=0))
figGI, axGI = plt.subplots(4, 3,figsize=(6,8),sharex='all')#,gridspec_kw=dict(hspace=0))
figII, axII = plt.subplots(4, 3,figsize=(6,8),sharex='all')#,gridspec_kw=dict(hspace=0))

#12 for HOD figsize=(6,24)
#4 for gamma_hat figsize=(6,8)

ip=-1
for param, value_range in gamma_hat_params.items():
    ip=ip+1
    iv=-1
    for val in value_range:
        iv=iv+1
        #command='hod_parameters_red.'+param+'=%.5e '%(val)
        #command+='hod_parameters_blue.'+param+'=%.5e '%(val)  #use -v when changing values
        command='radial_satellite_alignment_red.'+param+'=%.5e '%(val)
        command+='radial_satellite_alignment_blue.'+param+'=%.5e '%(val) #use -p when changing parameters
        print(command)
        test_sampler_run='cosmosis ../create_mock_with_HaloModel_IA.ini -p '+command
        test_sample_exit_status=os.system(test_sampler_run)
        plot_tomobins(ip,iv,'shear_cl',axall)
        plot_tomobins(ip,iv,'shear_cl_gi',axGI)
        plot_tomobins(ip,iv,'shear_cl_ii',axII)

    tidy_my_plot(axall,ip)
    tidy_my_plot(axGI,ip)
    tidy_my_plot(axII,ip)

figall.tight_layout()
figGI.tight_layout()
figII.tight_layout()

figall.savefig('Shear_All_varying_radial_calc_grid_params.png')
figGI.savefig('Shear_GI_varying_radial_calc_grid_params.png')
figII.savefig('Shear_II_varying_radial_calc_grid_params.png')



import matplotlib.pyplot as plt
import numpy as np

datadir='../output/HM_IA/'
def plot_cl(ax,shearcl,iz,jz,colour,linestyle):
    # Read in the power spectra files
    ellvals = np.loadtxt(datadir+shearcl+'/ell.txt')
    clvals = np.loadtxt(datadir+shearcl+'/bin_'+str(iz+1)+'_'+str(jz+1)+'.txt')
    ax.set_axis_on()
    ax.plot(ellvals, np.abs(clvals),label=shearcl,color=colour,linestyle=linestyle)

def plot_ratio_cl(ax,shearcl,cl_base,iz,jz,colour,linestyle):
    # Read in the power spectra files
    # Read in the power spectra files
    ellvals = np.loadtxt(datadir+shearcl+'/ell.txt')
    clvals = np.loadtxt(datadir+shearcl+'/bin_'+str(iz+1)+'_'+str(jz+1)+'.txt')
    ax.set_axis_on()
    ax.plot(ellvals,(clvals)/cl_base,label=shearcl,color=colour,linestyle=linestyle)

#First plot the Cl's
figone, ax = plt.subplots(5, 5,figsize=(10,10),sharey='all',sharex='all',gridspec_kw=dict(wspace=0,hspace=0))
for jz in range (5):
    for iz in range (5):
        if iz < jz:
            ax[iz,jz].set_axis_off()
        else:
            print (iz,jz)
            plot_cl(ax[iz,jz],'shear_cl',iz,jz,'black','solid')
            plot_cl(ax[iz,jz],'shear_cl_gg',iz,jz,'blue','dashed')
            plot_cl(ax[iz,jz],'shear_cl_gi',iz,jz,'cyan','solid')
            plot_cl(ax[iz,jz],'shear_cl_ii',iz,jz,'magenta','solid')
            ax[iz,jz].set_yscale('log')
            ax[iz,jz].set_xscale('log')
            if iz==4:
                ax[iz,jz].set_xlabel(r'$\ell$',fontsize=16)
            if jz==0:
                ax[iz,jz].set_ylabel(r'$C(\ell)$',fontsize=16)

ax[4,4].legend(loc='upper right',bbox_to_anchor=(1.0, 1.7))
plt.tight_layout()
figone.savefig('Cl_GG_GI_II_comparison.png')

# Now plot ratios to the NL matter power spectrum
figtwo, ax = plt.subplots(5, 5,figsize=(10,10),sharey='row',sharex='all',gridspec_kw=dict(wspace=0,hspace=0))
for jz in range (5):
    for iz in range (5):
        if iz < jz:
            ax[iz,jz].set_axis_off()
        else:
            print (iz,jz)
            cl_base = np.loadtxt(datadir+'shear_cl_gg/bin_'+str(iz+1)+'_'+str(jz+1)+'.txt')
            plot_ratio_cl(ax[iz,jz],'shear_cl',cl_base,iz,jz,'black','solid')
            plot_ratio_cl(ax[iz,jz],'shear_cl_gi',cl_base,iz,jz,'cyan','solid')
            plot_ratio_cl(ax[iz,jz],'shear_cl_ii',cl_base,iz,jz,'magenta','solid')
            ax[iz,jz].set_xscale('log')
            ax[iz,jz].hlines(y=1.0, xmin=0, xmax=3000, linestyle='dotted', color='grey')
            if iz==4:
                ax[iz,jz].set_xlabel(r'$\ell$',fontsize=16)
            if jz==0:
                ax[iz,jz].set_ylabel('$C(\\ell)/C_{\\rm GG}(\\ell)$',fontsize=16)

ax[4,4].legend(loc='upper right',bbox_to_anchor=(1.0, 1.7))
plt.tight_layout()
figtwo.savefig('Cl_GG_GI_II_ratio_comparison.png')



"""
A module for computing non-linear halo bias.
This module provides classes and functions to calculate non-linear halo bias using the Dark Emulator.
In future we might want to include other prescription of calculating the said quantity, namely the old Tinker05 and
an analytic prescription from Flamingo sims.
"""

import numpy as np
from collections import OrderedDict
from dark_emulator import darkemu
from scipy.interpolate import RegularGridInterpolator, interp1d
from scipy.optimize import curve_fit

from hmf._internals._cache import cached_quantity, parameter
from hmf._internals._framework import Framework


class NonLinearBias(Framework):
    """
    A class to compute the non-linear bias using the Dark Emulator.

    Parameters:
    -----------
    mass : array_like
        Array of halo masses.
    z_vec : array_like
        Array of redshifts.
    k_vec : array_like
        Array of wavenumbers.
    h0 : float
        Hubble parameter.
    sigma_8 : float
        Amplitude of matter fluctuations on 8 Mpc scales.
    A_s : float, optional
        Amplitude of the primordial power spectrum.
    omega_b : float
        Baryon density parameter.
    omega_c : float
        Cold dark matter density parameter.
    omega_lambda : float
        Dark energy density parameter.
    n_s : float
        Spectral index.
    w0 : float
        Dark energy equation of state parameter.
    z_dep : bool, optional
        If redshift dependence is to be evaluated in Bnl
    """

    def __init__(
        self,
        mass=None,
        z_vec=None,
        k_vec=None,
        h0=0.7,
        sigma_8=0.8,
        A_s=None,
        omega_b=0.05,
        omega_c=0.25,
        omega_lambda=0.7,
        n_s=1.0,
        w0=-1.0,
        z_dep=False,
    ):
        self.mass = mass
        self.z_vec = z_vec
        self.k_vec = k_vec
        self.h0 = h0
        self.sigma_8 = sigma_8
        self.A_s = A_s

        self.omega_b = omega_b
        self.omega_c = omega_c
        self.omega_lambda = omega_lambda
        self.n_s = n_s
        self.w0 = w0

        self.z_dep = z_dep

    @parameter('param')
    def mass(self, val):
        """
        Array of halo masses.

        :type: array_like
        """
        return val

    @parameter('param')
    def z_vec(self, val):
        """
        Array of redshifts.

        :type: array_like
        """
        return val

    @parameter('param')
    def k_vec(self, val):
        """
        Array of wavenumbers.

        :type: array_like
        """
        return val

    @parameter('param')
    def sigma_8(self, val):
        """
        Amplitude of matter fluctuations on 8 Mpc scales.

        :type: float
        """
        return val

    @parameter('param')
    def A_s(self, val):
        """
        Amplitude of the primordial power spectrum.

        :type: float
        """
        return val

    @parameter('param')
    def h0(self, val):
        """
        Hubble parameter.

        :type: float
        """
        return val

    @parameter('param')
    def omega_b(self, val):
        """
        Baryon density parameter.

        :type: float
        """
        return val

    @parameter('param')
    def omega_c(self, val):
        """
        Cold dark matter density parameter.

        :type: float
        """
        return val

    @parameter('param')
    def omega_lambda(self, val):
        """
        Dark energy density parameter.

        :type: float
        """
        return val

    @parameter('param')
    def n_s(self, val):
        """
        Spectral index.

        :type: float
        """
        return val

    @parameter('param')
    def w0(self, val):
        """
        Dark energy equation of state parameter.

        :type: float
        """
        return val

    @parameter('param')
    def z_dep(self, val):
        """
        If redshift dependence is to be evaluated in Bnl

        :type: bool
        """
        return val

    @cached_quantity
    def ombh2(self):
        """
        Return the physical baryon density parameter.

        Returns:
        --------
        array_like
            ombh2
        """
        return self.omega_b * self.h0**2.0

    @cached_quantity
    def omch2(self):
        """
        Return the physical cold dark matter density parameter.

        Returns:
        --------
        array_like
            omch2
        """
        return self.omega_c * self.h0**2.0

    @cached_quantity
    def emulator(self):
        """
        Initialize the Dark Emulator with the given cosmological parameters.

        Returns:
        --------
        darkemu.base_class
            An instance of the Dark Emulator.
        """
        emu = darkemu.base_class()
        if self.A_s is None and self.sigma_8 is not None:
            A_s_init = 2.1e-9

            cparam = self.test_cosmo(
                np.array(
                    [
                        self.ombh2,
                        self.omch2,
                        self.omega_lambda - 0.00064,
                        np.log(A_s_init * 1e10),
                        self.n_s,
                        self.w0,
                    ]
                )
            )
            emu.set_cosmology(cparam)

            sigma_8_init = emu.get_sigma8()
            scaling = self.sigma_8**2.0 / sigma_8_init**2.0
            A_s = A_s_init * scaling
            lnA = np.log(A_s * 1e10)
        elif self.A_s is not None:
            # We preffer A_s for DQ emulator!
            lnA = np.log(self.A_s * 1e10)
        else:
            raise ValueError('One of A_s or sigma_8 need to be specified!')

        cparam = self.test_cosmo(
            np.array(
                [self.ombh2, self.omch2, self.omega_lambda, lnA, self.n_s, self.w0]
            )
        )
        emu.set_cosmology(cparam)
        return emu

    @cached_quantity
    def bnl(self):
        """
        Compute the non-linear bias interpolation function.

        Returns:
        --------
        ndarray
            Interpolated non-linear bias values.
        """
        beta_interp_tmp = self.create_bnl_interpolation_function

        x = np.log10(self.mass)
        y = np.log10(self.mass)
        z = np.log10(self.k_vec)

        n, m, p = len(x), len(y), len(z)

        # Precompute all combinations of values (reordered for direct fill)
        Z = np.repeat(z, n * m)  # p varies fastest (outer axis first)
        X = np.tile(np.repeat(x, m), p)
        Y = np.tile(y, n * p)
        values = np.column_stack((X, Y, Z))

        # Precompute flat indices for direct assignment
        kk = np.repeat(np.arange(p), n * m)
        ii = np.tile(np.repeat(np.arange(n), m), p)
        jj = np.tile(np.arange(m), n * p)

        # Allocate and fill beta_interp efficiently
        if self.z_dep:
            beta_interp = np.zeros((len(self.z_vec), p, n, m))
            for i, _zi in enumerate(self.z_vec):
                beta_interp[i, kk, ii, jj] = beta_interp_tmp[i](values)
            return beta_interp
        else:
            beta_interp = np.zeros((p, n, m))
            beta_interp[kk, ii, jj] = beta_interp_tmp(values)
            return beta_interp[np.newaxis, :, :, :]

    def low_k_truncation(self, k, k_trunc):
        """
        Apply low-k truncation to the non-linear bias.

        Parameters:
        -----------
        k : array_like
            Wavenumber.
        k_trunc : float
            Truncation wavenumber.

        Returns:
        --------
        ndarray
            Truncation array.
        """
        return 1.0 / (1.0 + np.exp(-(10.0 * (np.log10(k) - np.log10(k_trunc)))))

    def high_k_truncation(self, k, k_trunc):
        """
        Apply high-k truncation to the non-linear bias.

        Parameters:
        -----------
        k : array_like
            Wavenumber.
        k_trunc : float
            Truncation wavenumber.

        Returns:
        --------
        ndarray
            Truncation array.
        """
        return 1.0 / (1.0 + np.exp(10.0 * (np.log10(k) - np.log10(k_trunc))))

    @property
    def minimum_halo_mass(self):
        """
        Compute the minimum halo mass for the set of cosmological parameters.

        Returns:
        --------
        tuple
            Minimum halo mass and corresponding wavenumber.
        """
        np_min = 200.0  # Minimum number of halo particles
        npart = 2048.0  # Cube root of number of simulation particles
        Lbox_HR = 1000.0  # Box size for high-resolution simulations [Mpc/h]
        Lbox_LR = 2000.0  # Box size for low-resolution simulations [Mpc/h] # noqa: F841

        Om_m = self.emulator.cosmo.get_Omega0()
        rhom = 2.77536627e11 * Om_m

        Mbox_HR = rhom * Lbox_HR**3.0
        mmin = Mbox_HR * np_min / npart**3.0

        vmin = Lbox_HR**3.0 * np_min / npart**3.0
        rmin = ((3.0 * vmin) / (4.0 * np.pi)) ** (1.0 / 3.0)

        return mmin, 2.0 * np.pi / rmin

    def rvir(self, mass):
        """
        Compute the virial radius for a given halo mass.

        Parameters:
        -----------
        mass : array_like
            Halo mass.

        Returns:
        --------
        ndarray
            Virial radius.
        """
        Om_m = self.emulator.cosmo.get_Omega0()
        rhom = 2.77536627e11 * Om_m
        return ((3.0 * mass) / (4.0 * np.pi * 200 * rhom)) ** (1.0 / 3.0)

    def hl_envelopes_idx(self, data, dmin=1, dmax=1):
        """
        Extract high and low envelope indices from a 1D data signal.

        Parameters:
        -----------
        data : array_like
            Data signal from which to extract high and low envelopes.
        dmin : int, optional
            Size of chunks for local minima.
        dmax : int, optional
            Size of chunks for local maxima.

        Returns:
        --------
        tuple
            Indices of high and low envelopes of the input signal.
        """
        # Find local minima indices
        lmin = (np.diff(np.sign(np.diff(data))) > 0).nonzero()[0] + 1
        # Find local maxima indices
        lmax = (np.diff(np.sign(np.diff(data))) < 0).nonzero()[0] + 1

        # Global min of dmin-chunks of local minima
        lmin = lmin[
            [i + np.argmin(data[lmin[i : i + dmin]]) for i in range(0, len(lmin), dmin)]
        ]
        # Global max of dmax-chunks of local maxima
        lmax = lmax[
            [i + np.argmax(data[lmax[i : i + dmax]]) for i in range(0, len(lmax), dmax)]
        ]

        return lmin, lmax

    def compute_bnl_darkquest(self, z, log10M1, log10M2, k, kmax):
        """
        Compute the non-linear bias using the Dark Emulator.

        Parameters:
        -----------
        z : float
            Redshift.
        log10M1 : array_like
            Log10 of the first halo mass.
        log10M2 : array_like
            Log10 of the second halo mass.
        k : array_like
            Wavenumber.
        kmax : float
            Maximum wavenumber.

        Returns:
        --------
        ndarray
            Non-linear bias values.
        """
        M1 = 10.0**log10M1
        M2 = 10.0**log10M2

        # Large 'linear' scale for linear halo bias [h/Mpc]
        klin = np.array([0.05])

        # Calculate beta_NL by looping over mass arrays
        beta_func = np.zeros((len(M1), len(M2), len(k)))

        # Linear power
        Pk_lin = self.emulator.get_pklin_from_z(k, z)
        Pk_klin = self.emulator.get_pklin_from_z(klin, z)

        # Calculate b01 for all M1
        b01 = np.zeros(len(M1))
        # b02 = np.zeros(len(M2))
        for iM, M0 in enumerate(M1):
            b01[iM] = np.sqrt(self.emulator.get_phh_mass(klin, M0, M0, z) / Pk_klin)

        for iM1, M01 in enumerate(M1):
            for iM2, M02 in enumerate(M2):
                if iM2 < iM1:
                    # Use symmetry to not double calculate
                    beta_func[iM1, iM2, :] = beta_func[iM2, iM1, :]
                else:
                    # Linear halo bias
                    b1 = b01[iM1]
                    b2 = b01[iM2]

                    # Halo-halo power spectrum
                    Pk_hh = self.emulator.get_phh_mass(k, M01, M02, z)

                    # rmax = max(self.rvir(M01), self.rvir(M02))
                    # kmax = 2.0*np.pi/rmax

                    # Create beta_NL
                    shot_noise = lambda x, a: a
                    popt, popc = curve_fit(
                        shot_noise,
                        k[(k > 100) & (k < 200)],
                        Pk_hh[(k > 100) & (k < 200)],
                    )
                    Pk_hh = Pk_hh - np.ones_like(k) * shot_noise(k, *popt)

                    beta_func[iM1, iM2, :] = Pk_hh / (b1 * b2 * Pk_lin) - 1.0

                    Pk_hh0 = self.emulator.get_phh_mass(klin, M01, M02, z)
                    Pk_hh0 = Pk_hh0 - np.ones_like(klin) * shot_noise(klin, *popt)
                    db = Pk_hh0 / (b1 * b2 * Pk_klin) - 1.0

                    lmin, lmax = self.hl_envelopes_idx(
                        np.abs(beta_func[iM1, iM2, :] + 1.0)
                    )
                    beta_func_interp = interp1d(
                        k[lmax],
                        np.abs(beta_func[iM1, iM2, lmax] + 1.0),
                        kind='quadratic',
                        bounds_error=False,
                        fill_value='extrapolate',
                    )
                    beta_func[iM1, iM2, :] = (
                        beta_func_interp(k) - 1.0
                    )  # * low_k_truncation(k, klin)
                    db = beta_func_interp(klin) - 1.0

                    # beta_func[iM1, iM2, :] = ((beta_func[iM1, iM2, :] + 1.0) * high_k_truncation(k, 30.0)/(db + 1.0) - 1.0) * low_k_truncation(k, klin)
                    # beta_func[iM1, iM2, :] = ((beta_func[iM1, iM2, :] + 1.0)/(db + 1.0) - 1.0) #* low_k_truncation(k, klin) * high_k_truncation(k, 30.0)#/(1.0+z))
                    beta_func[iM1, iM2, :] = (
                        (beta_func[iM1, iM2, :] - db)
                        * self.low_k_truncation(k, klin)
                        * self.high_k_truncation(k, 3.0 * kmax)
                    )

        return beta_func

    @cached_quantity
    def create_bnl_interpolation_function(self):
        """
        Create an interpolation function for the non-linear bias.

        Returns:
        --------
        RegularGridInterpolator
            Interpolation function for the non-linear bias.
        """
        lenM = 5
        lenk = 1000
        zc = self.z_vec.copy()

        Mmin, kmax = self.minimum_halo_mass
        M_up = np.log10(10.0**14.0)
        # M_lo = np.log10((10.0**12.0))
        M_lo = np.log10(Mmin)

        M = np.logspace(M_lo, M_up, lenM)
        k = np.logspace(-3.0, np.log10(200), lenk)

        if not self.z_dep:
            beta_func = self.compute_bnl_darkquest(
                0.01, np.log10(M), np.log10(M), k, kmax
            )
            beta_nl_interp_i = RegularGridInterpolator(
                [np.log10(M), np.log10(M), np.log10(k)],
                beta_func,
                fill_value=None,
                bounds_error=False,
                method='nearest',
            )

        if self.z_dep:
            beta_nl_interp_i = np.empty(len(self.z_vec), dtype=object)
            for i, zi in enumerate(zc):
                beta_func = self.compute_bnl_darkquest(
                    zi, np.log10(M), np.log10(M), k, kmax
                )
                beta_nl_interp_i[i] = RegularGridInterpolator(
                    [np.log10(M), np.log10(M), np.log10(k)],
                    beta_func,
                    fill_value=None,
                    bounds_error=False,
                    method='nearest',
                )
        return beta_nl_interp_i

    def test_cosmo(self, cparam_in):
        """
        Adjust cosmological parameters to be within the range of the Dark Emulator.

        Parameters:
        -----------
        cparam_in : array_like
            Input cosmological parameters.

        Returns:
        --------
        ndarray
            Adjusted cosmological parameters.
        """
        cparam_range = OrderedDict(
            [
                ['omegab', [0.0211375, 0.0233625]],
                ['omegac', [0.10782, 0.13178]],
                ['Omagede', [0.54752, 0.82128]],
                ['ln(10^10As)', [2.4752, 3.7128]],
                ['ns', [0.916275, 1.012725]],
                ['w', [-1.2, -0.8]],
            ]
        )

        cparam_in = cparam_in.reshape(1, 6)
        cparam_out = np.copy(cparam_in)

        for i, (_key, edges) in enumerate(cparam_range.items()):
            if cparam_in[0, i] < edges[0]:
                cparam_out[0, i] = edges[0]
            if cparam_in[0, i] > edges[1]:
                cparam_out[0, i] = edges[1]

        return cparam_out



import numpy as np
from abc import ABCMeta, abstractmethod

from hmf._internals._framework import Component, pluggable

# We name the classes here with lowercase in order to keep it simpler for the user.


@pluggable
class Poisson(Component, metaclass=ABCMeta):
    """
    Class to calculate the Poisson parameter for use in Pgg integrals.

    Can be either a scalar (P = poisson) or a power law (P = poisson x (M/M_0)^slope),
    by using the appropriate child class. Parent class is just an instance of an abstract class
    Further models can be added if necessary.

    Parameters:
    -----------
    mass : array_like
        Array of halo masses.

    """

    def __init__(self, mass, **model_parameters):
        self.mass = mass

        super().__init__(**model_parameters)

    @property
    @abstractmethod
    def poisson_func(self):  # pragma: no cover
        return


class constant(Poisson):
    """Constant Poisson parameter."""

    _defaults = {'poisson': 1.0}

    @property
    def poisson_func(self):
        return self.params['poisson'] * np.ones_like(self.mass)


class power_law(Poisson):
    """Power law mass dependent Poisson parameter."""

    _defaults = {'poisson': 1.0, 'pivot': None, 'slope': None}

    @property
    def poisson_func(self):
        poisson = self.params['poisson']
        M_0 = self.params['pivot']
        slope = self.params['slope']
        if M_0 is None or slope is None:
            raise ValueError(
                "pivot and slope must be provided for 'PowerLaw' poisson type."
            )
        return poisson * (self.mass / (10.0**M_0)) ** slope


# Add more if needed!



import pytest
import numpy as np
from scipy.integrate import quad
from scipy.interpolate import interp1d
from onepower import CosmologyBase, HaloModelIngredients
from onepower.hmi import SOVirial_Mead

# Helper tolerance for float comparisons
TOL = 1e-5


@pytest.fixture
def setup_data():
    k_vec = np.logspace(-4, 4, 100)
    z_vec = np.linspace(0.0, 3.0, 15)
    return k_vec, z_vec


@pytest.fixture
def ingredients(setup_data):
    k_vec, z_vec = setup_data
    return HaloModelIngredients(k_vec=k_vec, z_vec=z_vec)


@pytest.fixture
def cosmo():
    return CosmologyBase()


def test_SOVirial_Mead_properties():
    sov = SOVirial_Mead()
    dens = sov.halo_density(z=0)
    assert np.isclose(dens, sov.params['overdensity'] * sov.mean_density(0))
    assert sov.colossus_name == '200c'
    assert str(sov) == 'SOVirial'


def test_cosmology_base_properties(cosmo, setup_data):
    _, z_vec = setup_data
    assert np.allclose(cosmo.z_vec, z_vec)
    assert cosmo.h0 == 0.7
    assert cosmo.omega_c == 0.25
    assert cosmo.omega_b == 0.05
    assert cosmo.omega_m == 0.3
    assert cosmo.w0 == -1.0
    assert cosmo.wa == 0.0
    assert cosmo.n_s == 0.9
    assert cosmo.tcmb == 2.7255
    assert cosmo.m_nu == 0.06
    assert cosmo.sigma_8 == 0.8
    assert cosmo.log10T_AGN == 7.8


def test_cosmology_model_and_scale_factor(cosmo):
    cosmo_model = cosmo.cosmo_model
    assert hasattr(cosmo_model, 'H0')
    assert hasattr(cosmo_model, 'Om0')
    assert hasattr(cosmo_model, 'Ob0')
    assert hasattr(cosmo_model, 'Tcmb0')
    assert hasattr(cosmo_model, 'w0')
    assert hasattr(cosmo_model, 'wa')

    a = cosmo.scale_factor
    assert np.isclose(a[0], cosmo.cosmo_model.scale_factor(cosmo.z_vec[0]))
    assert len(a) == len(cosmo.z_vec)
    assert a[-1] < a[0]


@pytest.mark.parametrize('a,Om,Ode,Ok', [(0.5, 0.3, 0.7, 0), (1.0, 0.3, 0.7, 0)])
def test_Hubble_and_Omega_functions(cosmo, a, Om, Ode, Ok):
    H2 = cosmo._Hubble2(a, Om, Ode, Ok)
    Omega_m = cosmo._Omega_m(a, Om, Ode, Ok)
    AH = cosmo._AH(a, Om, Ode)
    assert H2 > 0
    assert 0 <= Omega_m <= 1
    assert np.isfinite(AH)


def test_mead_growth_functions(cosmo):
    g_fnc = cosmo.get_mead_growth_fnc
    assert isinstance(g_fnc, interp1d)
    a = np.linspace(0.001, 1, 10)
    g_vals = g_fnc(a)
    assert np.all(np.diff(g_vals) >= -1e-8)

    growth = cosmo.get_mead_growth
    assert growth.shape == cosmo.scale_factor.shape

    G = cosmo.get_mead_accumulated_growth
    assert G.shape == cosmo.scale_factor.shape
    assert np.all(np.diff(G) < 0)

    a0 = cosmo.scale_factor[0]
    val, err = quad(lambda a: cosmo.get_mead_growth_fnc(a) / a, 1e-4, a0)
    assert np.isclose(G[0], val + cosmo.get_mead_growth_fnc(1e-4), atol=1e-4)


@pytest.mark.parametrize(
    'x,y,p0,p1,p2,p3,expected',
    [
        (0.5, 0.5, 1, 2, 3, 4, 1 + 2 * (1 - 0.5) + 3 * (1 - 0.5) ** 2 + 4 * (1 - 0.5)),
        (
            1.0,
            0.0,
            0,
            -1,
            0.5,
            0,
            0 + -1 * (1 - 1.0) + 0.5 * (1 - 1.0) ** 2 + 0 * (1 - 0.0),
        ),
        (
            0.0,
            1.0,
            -1,
            0,
            0,
            1,
            -1 + 0 * (1 - 0.0) + 0 * (1 - 0.0) ** 2 + 1 * (1 - 1.0),
        ),
    ],
)
def test_f_Mead_function(cosmo, x, y, p0, p1, p2, p3, expected):
    result = cosmo.f_Mead(x, y, p0, p1, p2, p3)
    assert np.isclose(result, expected)


def test_dc_Mead_and_Dv_Mead_values(cosmo):
    dc = cosmo.dc_Mead
    Dv = cosmo.Dv_Mead
    assert dc.shape == cosmo.z_vec.shape
    assert Dv.shape == cosmo.z_vec.shape
    assert np.all((dc > 1.5) & (dc < 1.8))
    assert np.all(Dv > 100)
    assert np.all(Dv < 400)


def test_halo_model_ingredients_initialization(ingredients, setup_data):
    k_vec, _ = setup_data
    assert np.allclose(ingredients.k_vec, k_vec)
    assert ingredients.lnk_min == np.log(10 ** (-4.0))
    assert ingredients.lnk_max == np.log(10 ** (4.0))
    assert ingredients.dlnk == (np.log(10 ** (4.0)) - np.log(10 ** (-4.0))) / 100
    assert ingredients.Mmin == 9.0
    assert ingredients.Mmax == 16.0
    assert ingredients.dlog10m == 0.05
    assert ingredients.mdef_model == 'SOMean'
    assert ingredients.hmf_model == 'Tinker10'
    assert ingredients.bias_model == 'Tinker10'
    assert ingredients.halo_profile_model_dm == 'NFW'
    assert ingredients.halo_concentration_model_dm == 'Duffy08'
    assert ingredients.halo_profile_model_sat == 'NFW'
    assert ingredients.halo_concentration_model_sat == 'Duffy08'
    assert ingredients.transfer_model == 'CAMB'
    assert ingredients.transfer_params == {}
    assert ingredients.growth_model == 'CambGrowth'
    assert ingredients.growth_params == {}
    assert ingredients.norm_cen == 1.0
    assert ingredients.norm_sat == 1.0
    assert ingredients.eta_cen == 0.0
    assert ingredients.eta_sat == 0.0
    assert ingredients.overdensity == 200
    assert ingredients.delta_c == 1.686


def test_wrong_mead_option(ingredients, setup_data):
    with pytest.raises(ValueError):
        ingredients.update(mead_correction='something')
    ingredients.update(mead_correction=None)
    assert ingredients.mead_correction is None


def test_halo_model_ingredients_properties(ingredients, setup_data):
    _, z_vec = setup_data
    k_vec, _ = setup_data

    norm_c = ingredients._norm_c
    assert norm_c.shape == z_vec.shape

    norm_s = ingredients._norm_s
    assert norm_s.shape == z_vec.shape

    eta_c = ingredients._eta_c
    assert eta_c.shape == z_vec.shape

    eta_s = ingredients._eta_s
    assert eta_s.shape == z_vec.shape

    delta_c_mod = ingredients._delta_c_mod
    ingredients.update(mead_correction='feedback')
    delta_c_mod_mead = ingredients._delta_c_mod
    assert delta_c_mod.shape == z_vec.shape
    assert delta_c_mod_mead.shape == z_vec.shape
    assert np.allclose(delta_c_mod_mead, ingredients.dc_Mead)
    ingredients.update(mead_correction=None)

    mdef_mod = ingredients._mdef_mod
    ingredients.update(mead_correction='feedback')
    mdef_mod_mead = ingredients._mdef_mod
    assert isinstance(mdef_mod, str)
    assert isinstance(mdef_mod_mead, type(SOVirial_Mead))
    ingredients.update(mead_correction=None)

    hmf_mod = ingredients._hmf_mod
    ingredients.update(mead_correction='feedback')
    hmf_mod_mead = ingredients._hmf_mod
    assert isinstance(hmf_mod, str)
    assert isinstance(hmf_mod_mead, str)
    assert hmf_mod_mead == 'ST'
    ingredients.update(mead_correction=None)

    bias_mod = ingredients._bias_mod
    ingredients.update(mead_correction='feedback')
    bias_mod_mead = ingredients._bias_mod
    assert isinstance(bias_mod, str)
    assert isinstance(bias_mod_mead, str)
    assert bias_mod_mead == 'ST99'
    ingredients.update(mead_correction=None)

    halo_concentration_mod_dm = ingredients._halo_concentration_mod_dm
    assert callable(halo_concentration_mod_dm)
    ingredients.update(halo_concentration_model_dm='diemer19')
    halo_concentration_mod_dm = ingredients._halo_concentration_mod_dm
    assert callable(halo_concentration_mod_dm)
    ingredients.update(
        halo_concentration_model_dm='Duffy08', mead_correction='feedback'
    )
    halo_concentration_mod_dm_mead = ingredients._halo_concentration_mod_dm
    assert callable(halo_concentration_mod_dm_mead)
    ingredients.update(mead_correction=None)

    halo_concentration_mod_sat = ingredients._halo_concentration_mod_sat
    assert callable(halo_concentration_mod_sat)
    ingredients.update(halo_concentration_model_sat='diemer19')
    halo_concentration_mod_sat = ingredients._halo_concentration_mod_sat
    assert callable(halo_concentration_mod_sat)
    ingredients.update(halo_concentration_model_sat='Duffy08')


def test_halo_model_ingredients_miscellaneous(ingredients, setup_data):
    _, z_vec = setup_data
    k_vec, _ = setup_data

    mdef_params = ingredients.mdef_params
    assert isinstance(mdef_params, list)

    halo_profile_params = ingredients.halo_profile_params
    assert isinstance(halo_profile_params, dict)

    scale_factor = ingredients.scale_factor
    assert isinstance(scale_factor, np.ndarray)

    disable_mass_conversion = ingredients.disable_mass_conversion
    assert isinstance(disable_mass_conversion, bool)
    ingredients.update(mead_correction='feedback')
    disable_mass_conversion = ingredients.disable_mass_conversion
    assert isinstance(disable_mass_conversion, bool)
    ingredients.update(mead_correction=None)

    K = ingredients.K
    assert isinstance(K, np.ndarray)
    assert np.allclose(K, np.zeros_like(z_vec))
    ingredients.update(mead_correction='feedback')
    K = ingredients.K
    assert isinstance(K, np.ndarray)
    ingredients.update(mead_correction='nofeedback')
    K = ingredients.K
    assert isinstance(K, np.ndarray)
    ingredients.update(mead_correction=None)

    hmf_generator = ingredients._hmf_generator
    assert isinstance(hmf_generator, tuple)
    assert len(hmf_generator) == 2
    ingredients.update(mead_correction='feedback')
    hmf_generator = ingredients._hmf_generator
    assert isinstance(hmf_generator, tuple)
    assert len(hmf_generator) == 2
    ingredients.update(mead_correction=None)

    hmf_cen = ingredients._hmf_cen
    assert isinstance(hmf_cen, list)
    assert len(hmf_cen) == z_vec.size

    hmf_sat = ingredients._hmf_sat
    assert isinstance(hmf_sat, list)
    assert len(hmf_sat) == z_vec.size

    mass = ingredients.mass
    assert isinstance(mass, np.ndarray)

    power = ingredients.power
    assert isinstance(power, np.ndarray)
    assert power.shape == (z_vec.size, k_vec.size)

    nonlinear_power = ingredients.nonlinear_power
    assert isinstance(nonlinear_power, np.ndarray)
    assert nonlinear_power.shape == (z_vec.size, k_vec.size)

    kh = ingredients.kh
    assert isinstance(kh, np.ndarray)
    assert kh.size == k_vec.size

    halo_overdensity_mean = ingredients.halo_overdensity_mean
    assert isinstance(halo_overdensity_mean, np.ndarray)
    assert halo_overdensity_mean.shape[0] == z_vec.size

    nu = ingredients.nu
    assert isinstance(nu, np.ndarray)
    assert nu.shape[0] == z_vec.size

    dndlnm = ingredients.dndlnm
    assert isinstance(dndlnm, np.ndarray)
    assert dndlnm.shape[0] == z_vec.size

    mean_density0 = ingredients.mean_density0
    assert isinstance(mean_density0, np.ndarray)
    assert mean_density0.size == z_vec.size

    mean_density_z = ingredients.mean_density_z
    assert isinstance(mean_density_z, np.ndarray)
    assert mean_density_z.size == z_vec.size

    rho_halo = ingredients.rho_halo
    assert isinstance(rho_halo, np.ndarray)
    assert rho_halo.size == z_vec.size

    halo_bias = ingredients.halo_bias
    assert isinstance(halo_bias, np.ndarray)
    assert halo_bias.shape[0] == z_vec.size

    neff = ingredients.neff
    assert isinstance(neff, np.ndarray)
    assert neff.shape[0] == z_vec.size

    sigma8_z = ingredients.sigma8_z
    assert isinstance(sigma8_z, np.ndarray)
    assert sigma8_z.shape[0] == z_vec.size

    fnu = ingredients.fnu
    assert isinstance(fnu, np.ndarray)
    assert fnu.shape[0] == z_vec.size

    conc_cen = ingredients.conc_cen
    assert isinstance(conc_cen, np.ndarray)
    assert conc_cen.shape[0] == z_vec.size

    nfw_cen = ingredients.nfw_cen
    assert isinstance(nfw_cen, np.ndarray)
    assert nfw_cen.shape[0] == z_vec.size

    u_dm = ingredients.u_dm
    assert isinstance(u_dm, np.ndarray)
    assert u_dm.shape[0] == z_vec.size
    assert u_dm[0, 0, 0] == 1.0

    r_s_cen = ingredients.r_s_cen
    assert isinstance(r_s_cen, np.ndarray)
    assert r_s_cen.shape[0] == z_vec.size

    rvir_cen = ingredients.rvir_cen
    assert isinstance(rvir_cen, np.ndarray)
    assert rvir_cen.shape[0] == z_vec.size

    conc_sat = ingredients.conc_sat
    assert isinstance(conc_sat, np.ndarray)
    assert conc_sat.shape[0] == z_vec.size

    nfw_sat = ingredients.nfw_sat
    assert isinstance(nfw_sat, np.ndarray)
    assert nfw_sat.shape[0] == z_vec.size

    u_sat = ingredients.u_sat
    assert isinstance(u_sat, np.ndarray)
    assert u_sat.shape[0] == z_vec.size
    assert u_sat[0, 0, 0] == 1.0

    r_s_sat = ingredients.r_s_sat
    assert isinstance(r_s_sat, np.ndarray)
    assert r_s_sat.shape[0] == z_vec.size

    rvir_sat = ingredients.rvir_sat
    assert isinstance(rvir_sat, np.ndarray)
    assert rvir_sat.shape[0] == z_vec.size

    growth_factor = ingredients.growth_factor
    assert isinstance(growth_factor, np.ndarray)
    assert growth_factor.shape[0] == z_vec.size



import numbers
import numpy as np
from cosmosis.datablock import names, option_section
from scipy.interpolate import interp1d

from onepower.add import UpsampledSpectra
from onepower.bnl import NonLinearBias
from onepower.pk import Spectra

cosmo_params = names.cosmological_parameters

parameters_models = {
    'Zheng': ['log10_Mmin', 'sigma', 'log10_M0', 'log10_M1', 'alpha'],
    'Zhai': ['log10_Mmin', 'sigma', 'log10_Msat', 'log10_Mcut', 'alpha'],
    'Cacciato': [
        'log10_obs_norm_c',
        'log10_m_ch',
        'g1',
        'g2',
        'sigma_log10_O_c',
        'norm_s',
        'pivot',
        'alpha_s',
        'beta_s',
        'b0',
        'b1',
        'b2',
    ],
}

poisson_parameters = {
    'constant': ['poisson'],
    'power_law': ['poisson', 'pivot', 'slope'],
}

# Mapping of use_mead values to mead_correction values
mead_correction_map = {
    'mead2020': 'nofeedback',
    'mead2020_feedback': 'feedback',
    'fit_feedback': 'fit',
    # Add more mappings here if needed
}


def get_string_or_none(cosmosis_block, section, name, default):
    """
    A helper function to return a number or None explicitly from config files
    or return None if no value is present.
    """
    if cosmosis_block.has_value(section, name):
        test_param = cosmosis_block.get(section, name)
        if isinstance(test_param, numbers.Number):
            param = cosmosis_block.get_double(section, name, default)
        if isinstance(test_param, str):
            str_in = cosmosis_block.get_string(section, name)
            if str_in == 'None':
                param = None
    else:
        try:
            param = cosmosis_block.get_double(section, name, default)
        except ValueError:
            param = None

    if not isinstance(param, (numbers.Number | type(None))):
        raise ValueError(
            f'Parameter {name} is not an instance of a number or NoneType!'
        )

    return param


def interpolate_in_z(input_grid, z_in, z_out, axis=0):
    """
    Interpolation in redshift
    Default redshift axis is the first one.
    """
    f_interp = interp1d(z_in, input_grid, axis=axis)
    return f_interp(z_out)


def log_linear_interpolation_k(power_in, k_in, k_out, axis=1, kind='linear'):
    """
    log-linear interpolation for power spectra. This works well for extrapolating to higher k.
    Ideally we want to have a different routine for interpolation (spline) and extrapolation (log-linear)
    """
    power_interp = interp1d(
        np.log(k_in), np.log(power_in), axis=axis, kind=kind, fill_value='extrapolate'
    )
    return np.exp(power_interp(np.log(k_out)))


def get_linear_power_spectrum(block, z_vec):
    """
    Reads in linear matter power spectrum and downsamples
    """
    k_vec = block['matter_power_lin', 'k_h']
    z_pl = block['matter_power_lin', 'z']
    matter_power_lin = block['matter_power_lin', 'p_k']
    return k_vec, interpolate_in_z(matter_power_lin, z_pl, z_vec)


def get_nonlinear_power_spectrum(block, z_vec):
    """
    Reads in the non-linear matter power specturm and downsamples
    """
    k_nl = block['matter_power_nl', 'k_h']
    z_nl = block['matter_power_nl', 'z']
    matter_power_nl = block['matter_power_nl', 'p_k']
    return k_nl, interpolate_in_z(matter_power_nl, z_nl, z_vec)


def setup_hod_settings(options, suffix):
    """Setup HOD settings based on options."""
    hod_settings = {}
    if options.has_value(option_section, f'observables_file_{suffix}'):
        hod_settings['observables_file'] = options.get_string(
            option_section, f'observables_file_{suffix}'
        )
    else:
        hod_settings['observables_file'] = None
        hod_settings['obs_min'] = np.asarray(
            [options[option_section, f'log10_obs_min_{suffix}']]
        ).flatten()
        hod_settings['obs_max'] = np.asarray(
            [options[option_section, f'log10_obs_max_{suffix}']]
        ).flatten()
        hod_settings['zmin'] = np.asarray(
            [options[option_section, f'zmin_{suffix}']]
        ).flatten()
        hod_settings['zmax'] = np.asarray(
            [options[option_section, f'zmax_{suffix}']]
        ).flatten()
        hod_settings['nz'] = options[option_section, f'nz_{suffix}']
    hod_settings['nobs'] = options[option_section, f'nobs_{suffix}']
    hod_settings['observable_h_unit'] = options.get_string(
        option_section, 'observable_h_unit', default='1/h^2'
    ).lower()
    return hod_settings


def setup_hod_settings_mm(hod_settings):
    """Setup matter-matter HOD settings based on options."""
    hod_settings_mm = hod_settings.copy()
    if hod_settings_mm['observables_file'] is None:
        hod_settings_mm['obs_min'] = np.array([hod_settings['obs_min'].min()])
        hod_settings_mm['obs_max'] = np.array([hod_settings['obs_max'].max()])
        hod_settings_mm['zmin'] = np.array([hod_settings['zmin'].min()])
        hod_settings_mm['zmax'] = np.array([hod_settings['zmax'].max()])
        hod_settings_mm['nz'] = 15
    hod_settings_mm['nobs'] = 100
    return hod_settings_mm


def setup_hmf_config(options):
    """Setup function to parse options and return configuration for hmf."""
    config_hmf = {
        'log_mass_min': options.get_double(option_section, 'log_mass_min', default=9.0),
        'log_mass_max': options.get_double(
            option_section, 'log_mass_max', default=16.0
        ),
        'dlog10m': (
            options.get_double(option_section, 'log_mass_max', default=16.0)
            - options.get_double(option_section, 'log_mass_min', default=9.0)
        )
        / options.get_int(option_section, 'nmass', default=200),
        'z_vec': np.linspace(
            options.get_double(option_section, 'zmin_hmf'),
            options.get_double(option_section, 'zmax_hmf'),
            options.get_int(option_section, 'nz_hmf', default=15),
        ),
        'nk': options.get_int(option_section, 'nk', default=100),
        'profile_cen': options.get_string(option_section, 'profile_cen', default='NFW'),
        'profile_sat': options.get_string(option_section, 'profile_sat', default='NFW'),
        'profile_value_name': options.get_string(
            option_section, 'profile_value_name', default='profile_parameters'
        ),
        'hmf_model': options.get_string(
            option_section, 'hmf_model', default='Tinker10'
        ),
        'mdef_model': options.get_string(
            option_section, 'mdef_model', default='SOMean'
        ),
        'overdensity': options.get_double(option_section, 'overdensity', default=200.0),
        'cm_model_cen': options.get_string(
            option_section, 'cm_model_cen', default='Duffy08'
        ),
        'cm_model_sat': options.get_string(
            option_section, 'cm_model_sat', default='Duffy08'
        ),
        'delta_c': options.get_double(option_section, 'delta_c', default=1.686),
        'bias_model': options.get_string(
            option_section, 'bias_model', default='Tinker10'
        ),
        'lnk_min': np.log(1e-8),
        'lnk_max': np.log(1e8),
        'dlnk': 0.05,  # 0.001
    }
    return config_hmf


def setup_pipeline_parameters(options):
    """Setup pipeline parameters."""
    p_mm = options.get_bool(option_section, 'p_mm', default=False)
    p_gg = options.get_bool(option_section, 'p_gg', default=False)
    p_gm = options.get_bool(option_section, 'p_gm', default=False)
    p_gI = options.get_bool(option_section, 'p_gI', default=False)
    p_mI = options.get_bool(option_section, 'p_mI', default=False)
    p_II = options.get_bool(option_section, 'p_II', default=False)
    # If True, calculate the response of the halo model for the requested power spectra compared to matter power
    # multiplies this to input non-linear matter power spectra.
    response = options.get_bool(option_section, 'response', default=False)
    # If true, use the IA formalism of Fortuna et al. 2021: Truncated NLA at high k + 1-halo term
    fortuna = options.get_bool(option_section, 'fortuna', default=False)
    # If True, uses beta_nl
    bnl = options.get_bool(option_section, 'bnl', default=False)
    split_ia = options.get_bool(option_section, 'split_ia', default=False)

    matter = p_mm
    galaxy = p_gg or p_gm
    alignment = p_gI or p_mI or p_II

    # Fortuna introduces a truncation of the 1-halo term at large scales to avoid the halo exclusion problem
    # and a truncation of the NLA 2-halo term at small scales to avoid double-counting of the 1-halo term
    # The user can change these values.
    one_halo_ktrunc_ia = get_string_or_none(
        options, option_section, 'one_halo_ktrunc_ia', default=4.0
    )  # h/Mpc or None
    two_halo_ktrunc_ia = get_string_or_none(
        options, option_section, 'two_halo_ktrunc_ia', default=6.0
    )  # h/Mpc or None
    # General truncation of non-IA terms:
    one_halo_ktrunc = get_string_or_none(
        options, option_section, 'one_halo_ktrunc', default=0.1
    )  # h/Mpc or None
    two_halo_ktrunc = get_string_or_none(
        options, option_section, 'two_halo_ktrunc', default=2.0
    )  # h/Mpc or None

    # Additional parameters
    mead_correction = mead_correction_map.get(
        options.get_string(option_section, 'mead2020_correction', default='None'), None
    )
    dewiggle = options.get_bool(option_section, 'dewiggle', default=False)
    point_mass = options.get_bool(option_section, 'point_mass', default=False)
    poisson_type = options.get_string(option_section, 'poisson_type', default='')

    return (
        p_mm,
        p_gg,
        p_gm,
        p_gI,
        p_mI,
        p_II,
        response,
        fortuna,
        matter,
        galaxy,
        bnl,
        alignment,
        split_ia,
        one_halo_ktrunc,
        two_halo_ktrunc,
        one_halo_ktrunc_ia,
        two_halo_ktrunc_ia,
        mead_correction,
        dewiggle,
        point_mass,
        poisson_type,
    )


def setup_hod(options, alignment, split_ia):
    hod_section_name = options.get_string(option_section, 'hod_section_name')
    hod_values_name = options.get_string(
        option_section, 'values_name', default='hod_parameters'
    ).lower()

    population_name = options.get_string(
        option_section, 'output_suffix', default=''
    ).lower()
    pop_name = f'_{population_name}' if population_name else ''

    if alignment:
        central_IA = options.get_string(
            option_section, 'central_IA_depends_on', default='halo_mass'
        )
        satellite_IA = options.get_string(
            option_section, 'satellite_IA_depends_on', default='halo_mass'
        )
        if split_ia:
            hod_section_name_ia_1 = options.get_string(
                option_section, 'hod_section_name_ia_1'
            )
            population_name_ia_1 = options.get_string(
                option_section, 'output_suffix_ia_1', default=''
            ).lower()
            pop_name_ia_1 = f'_{population_name_ia_1}' if population_name_ia_1 else ''

            hod_section_name_ia_2 = options.get_string(
                option_section, 'hod_section_name_ia_2'
            )
            population_name_ia_2 = options.get_string(
                option_section, 'output_suffix_ia_2', default=''
            ).lower()
            pop_name_ia_2 = f'_{population_name_ia_2}' if population_name_ia_2 else ''
        else:
            hod_section_name_ia_1 = options.get_string(
                option_section, 'hod_section_name_ia'
            )
            population_name_ia_1 = options.get_string(
                option_section, 'output_suffix_ia', default=''
            ).lower()
            pop_name_ia_1 = f'_{population_name_ia_1}' if population_name_ia_1 else ''
            hod_section_name_ia_2 = None
            pop_name_ia_2 = None
    else:
        hod_section_name_ia_1 = None
        pop_name_ia_1 = ''
        hod_section_name_ia_2 = None
        pop_name_ia_2 = ''
        central_IA = None
        satellite_IA = None

    hod_model = options.get_string(option_section, 'hod_model', default='Cacciato')
    hod_settings = setup_hod_settings(options, 'hod')
    hod_settings_mm = setup_hod_settings_mm(hod_settings)
    nbins = len(hod_settings['obs_min'])
    if alignment:
        if split_ia:
            hod_settings_ia_1 = setup_hod_settings(options, 'ia_1')
            hod_settings_ia_2 = setup_hod_settings(options, 'ia_2')
        else:
            hod_settings_ia_1 = setup_hod_settings(options, 'ia')
            hod_settings_ia_2 = {}
    else:
        hod_settings_ia_1 = {}
        hod_settings_ia_2 = {}
    if options.get_bool(option_section, 'save_observable', default=True):
        obs_settings = setup_hod_settings(options, 'smf')
        obs_settings['save_observable'] = options.get_bool(
            option_section, 'save_observable', default=True
        )
        obs_settings['observable_section_name'] = options.get_string(
            option_section, 'observable_section_name', default='stellar_mass_function'
        ).lower()
    else:
        obs_settings = {}

    return (
        hod_section_name,
        hod_values_name,
        pop_name,
        hod_model,
        hod_settings,
        hod_settings_mm,
        obs_settings,
        nbins,
        hod_settings_ia_1,
        hod_settings_ia_2,
        hod_section_name_ia_1,
        hod_section_name_ia_2,
        pop_name_ia_1,
        pop_name_ia_2,
        central_IA,
        satellite_IA,
    )


def save_matter_results(block, power, z_vec, k_vec):
    """Save matter results to the block."""
    mass = power.mass
    mean_density0 = power.mean_density0
    mean_density_z = power.mean_density_z
    rho_crit = power.mean_density0 / block[cosmo_params, 'omega_m']
    rho_halo = power.rho_halo

    dndlnm = power.dndlnm
    halo_bias = power.halo_bias
    nu = power.nu
    neff = power.neff
    sigma8_z = power.sigma8_z
    fnu = power.fnu

    u_dm_cen = power.u_dm
    u_dm_sat = power.u_sat

    conc_cen = power.conc_cen
    conc_sat = power.conc_sat
    r_s_cen = power.r_s_cen
    r_s_sat = power.r_s_sat

    rvir_cen = power.rvir_cen
    rvir_sat = power.rvir_sat

    # TODO: Clean these up. Put more of them into the same folder
    block.put_grid('concentration_m', 'z', z_vec, 'm_h', mass, 'c', conc_cen)
    block.put_grid('concentration_sat', 'z', z_vec, 'm_h', mass, 'c', conc_sat)
    block.put_grid('nfw_scale_radius_m', 'z', z_vec, 'm_h', mass, 'rs', r_s_cen)
    block.put_grid('nfw_scale_radius_sat', 'z', z_vec, 'm_h', mass, 'rs', r_s_sat)

    block.put_double_array_1d('virial_radius', 'm_h', mass)
    # rvir doesn't change with z, hence no z-dimension
    block.put_double_array_1d('virial_radius', 'rvir_m', rvir_cen[0])
    block.put_double_array_1d('virial_radius', 'rvir_sat', rvir_sat[0])

    block.put_double_array_1d('fourier_nfw_profile', 'z', z_vec)
    block.put_double_array_1d('fourier_nfw_profile', 'm_h', mass)
    block.put_double_array_1d('fourier_nfw_profile', 'k_h', k_vec)
    block.put_double_array_nd('fourier_nfw_profile', 'ukm', u_dm_cen)
    block.put_double_array_nd('fourier_nfw_profile', 'uksat', u_dm_sat)

    # Density
    block['density', 'mean_density0'] = mean_density0
    block['density', 'rho_crit'] = rho_crit
    block.put_double_array_1d('density', 'mean_density_z', mean_density_z)
    block.put_double_array_1d('density', 'rho_halo', rho_halo)
    block.put_double_array_1d('density', 'z', z_vec)

    # Halo mass function
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'dndlnmh', dndlnm)
    block.put_grid('hmf', 'z', z_vec, 'm_h', mass, 'nu', nu)
    block.put_double_array_1d('hmf', 'neff', neff)
    block.put_double_array_1d('hmf', 'sigma8_z', sigma8_z)

    # Linear halo bias
    block.put_grid('halobias', 'z', z_vec, 'm_h', mass, 'b_hb', halo_bias)

    # Fraction of neutrinos to total matter, f_nu = Ω_nu /Ω_m
    block[cosmo_params, 'fnu'] = fnu


def save_hod_results(block, power, z_vec, hod_section_name, hod_settings):
    """Save HOD results to the block."""
    mass = power.mass
    N_cen = power.hod.hod_cen
    N_sat = power.hod.hod_sat
    N_tot = power.hod.hod
    numdens_cen = power.hod.number_density_cen
    numdens_sat = power.hod.number_density_sat
    numdens_tot = power.hod.number_density
    fraction_c = power.hod.f_c
    fraction_s = power.hod.f_s
    mass_avg = power.mass_avg
    f_star = power.fstar

    hod_bins = N_cen.shape[0]
    block.put_int(hod_section_name, 'nbins', hod_bins)
    for nb in range(hod_bins):
        suffix = f'_{nb + 1}' if hod_bins != 1 else ''
        block.put_grid(
            hod_section_name,
            f'z{suffix}',
            z_vec,
            f'mass{suffix}',
            mass,
            f'N_sat{suffix}',
            N_sat[nb],
        )
        block.put_grid(
            hod_section_name,
            f'z{suffix}',
            z_vec,
            f'mass{suffix}',
            mass,
            f'N_cen{suffix}',
            N_cen[nb],
        )
        block.put_grid(
            hod_section_name,
            f'z{suffix}',
            z_vec,
            f'mass{suffix}',
            mass,
            f'N_tot{suffix}',
            N_tot[nb],
        )
        block.put_grid(
            hod_section_name,
            f'z{suffix}',
            z_vec,
            f'mass{suffix}',
            mass,
            f'f_star{suffix}',
            f_star[nb],
        )
        block.put_double_array_1d(
            hod_section_name, f'number_density_cen{suffix}', numdens_cen[nb]
        )
        block.put_double_array_1d(
            hod_section_name, f'number_density_sat{suffix}', numdens_sat[nb]
        )
        block.put_double_array_1d(
            hod_section_name, f'number_density_tot{suffix}', numdens_tot[nb]
        )
        block.put_double_array_1d(
            hod_section_name, f'central_fraction{suffix}', fraction_c[nb]
        )
        block.put_double_array_1d(
            hod_section_name, f'satellite_fraction{suffix}', fraction_s[nb]
        )
        block.put_double_array_1d(
            hod_section_name, f'average_halo_mass{suffix}', mass_avg[nb]
        )
    return hod_bins


def save_obs_results(block, power, observable_section_name, obs_settings):
    obs_func = power.obs_func
    obs_func_c = power.obs_func_cen
    obs_func_s = power.obs_func_sat
    obs_z = power.obs_func_z
    obs_range = power.obs_func_obs
    obs_bins = obs_range.shape[0]

    block.put(observable_section_name, 'obs_func_definition', 'obs_func * obs * ln(10)')
    for nb in range(obs_bins):
        suffix_obs = f'_{nb + 1}'
        if np.all(
            np.array(
                [
                    obs_settings['obs_min'].size,
                    obs_settings['obs_max'].size,
                    obs_settings['zmin'].size,
                    obs_settings['zmax'].size,
                    obs_settings['nz'],
                ]
            )
            == 1
        ):
            block.put_double_array_1d(
                observable_section_name, f'obs_val{suffix_obs}', np.squeeze(obs_range)
            )
            block.put_double_array_1d(
                observable_section_name, f'obs_func{suffix_obs}', np.squeeze(obs_func)
            )
            block.put_double_array_1d(
                observable_section_name,
                f'obs_func_c{suffix_obs}',
                np.squeeze(obs_func_c),
            )
            block.put_double_array_1d(
                observable_section_name,
                f'obs_func_s{suffix_obs}',
                np.squeeze(obs_func_s),
            )
        else:
            block.put_grid(
                observable_section_name,
                f'z_bin{suffix_obs}',
                obs_z[nb],
                f'obs_val{suffix_obs}',
                obs_range[nb, 0, :],
                f'obs_func{suffix_obs}',
                obs_func[nb],
            )
            block.put_grid(
                observable_section_name,
                f'z_bin{suffix_obs}',
                obs_z[nb],
                f'obs_val{suffix_obs}',
                obs_range[nb, 0, :],
                f'obs_func_c{suffix_obs}',
                obs_func_c[nb],
            )
            block.put_grid(
                observable_section_name,
                f'z_bin{suffix_obs}',
                obs_z[nb],
                f'obs_val{suffix_obs}',
                obs_range[nb, 0, :],
                f'obs_func_s{suffix_obs}',
                obs_func_s[nb],
            )


def save_pk_to_grid(block, z_vec, k_vec, base_name, suffix, pk_1h, pk_2h, pk_tot):
    """Save P(k) to the block."""
    section_name = f'{base_name}{suffix}'
    block.put_grid(section_name, 'z', z_vec, 'k_h', k_vec, 'p_k_1h', pk_1h)
    block.put_grid(section_name, 'z', z_vec, 'k_h', k_vec, 'p_k_2h', pk_2h)
    block.put_grid(section_name, 'z', z_vec, 'k_h', k_vec, 'p_k', pk_tot)


def setup(options):
    (
        p_mm,
        p_gg,
        p_gm,
        p_gI,
        p_mI,
        p_II,
        response,
        fortuna,
        matter,
        galaxy,
        bnl,
        alignment,
        split_ia,
        one_halo_ktrunc,
        two_halo_ktrunc,
        one_halo_ktrunc_ia,
        two_halo_ktrunc_ia,
        mead_correction,
        dewiggle,
        point_mass,
        poisson_type,
    ) = setup_pipeline_parameters(options)

    (
        hod_section_name,
        hod_values_name,
        pop_name,
        hod_model,
        hod_settings,
        hod_settings_mm,
        obs_settings,
        nbins,
        hod_settings_ia_1,
        hod_settings_ia_2,
        hod_section_name_ia_1,
        hod_section_name_ia_2,
        pop_name_ia_1,
        pop_name_ia_2,
        central_IA,
        satellite_IA,
    ) = setup_hod(options, alignment, split_ia)

    # hmf config
    config_hmf = setup_hmf_config(options)

    if bnl:
        cached_bnl = {
            'num_calls': 0,
            'cached_bnl': None,
            'update_bnl': options[option_section, 'update_bnl'],
        }
    else:
        cached_bnl = None

    config_hmf['power'] = Spectra()

    return (
        p_mm,
        p_gg,
        p_gm,
        p_gI,
        p_mI,
        p_II,
        response,
        fortuna,
        matter,
        galaxy,
        bnl,
        alignment,
        split_ia,
        one_halo_ktrunc,
        two_halo_ktrunc,
        one_halo_ktrunc_ia,
        two_halo_ktrunc_ia,
        hod_section_name,
        mead_correction,
        dewiggle,
        point_mass,
        poisson_type,
        pop_name,
        hod_model,
        hod_settings,
        hod_settings_mm,
        obs_settings,
        hod_values_name,
        config_hmf,
        cached_bnl,
        central_IA,
        satellite_IA,
        nbins,
        hod_settings_ia_1,
        hod_settings_ia_2,
        hod_section_name_ia_1,
        hod_section_name_ia_2,
        pop_name_ia_1,
        pop_name_ia_2,
    )


def execute(block, config):
    """Execute function to compute power spectra based on configuration."""
    (
        p_mm,
        p_gg,
        p_gm,
        p_gI,
        p_mI,
        p_II,
        response,
        fortuna,
        matter,
        galaxy,
        bnl,
        alignment,
        split_ia,
        one_halo_ktrunc,
        two_halo_ktrunc,
        one_halo_ktrunc_ia,
        two_halo_ktrunc_ia,
        hod_section_name,
        mead_correction,
        dewiggle,
        point_mass,
        poisson_type,
        pop_name,
        hod_model,
        hod_settings,
        hod_settings_mm,
        obs_settings,
        hod_values_name,
        config_hmf,
        cached_bnl,
        central_IA,
        satellite_IA,
        nbins,
        hod_settings_ia_1,
        hod_settings_ia_2,
        hod_section_name_ia_1,
        hod_section_name_ia_2,
        pop_name_ia_1,
        pop_name_ia_2,
    ) = config

    hod_params = {}

    # Power spectrum transfer function used to update the transfer function in hmf
    transfer_k = block['matter_power_transfer_func', 'k_h']
    transfer_func = block['matter_power_transfer_func', 't_k']
    growth_z = block['growth_parameters', 'z']
    growth_func = block['growth_parameters', 'd_z']

    z_vec = config_hmf['z_vec']

    # Load the linear power spectrum and growth factor
    k_vec_original, plin_original = get_linear_power_spectrum(block, z_vec)
    k_vec = np.logspace(
        np.log10(k_vec_original[0]), np.log10(k_vec_original[-1]), num=config_hmf['nk']
    )  # , endpoint=False)
    z_vec_original = block.get_double_array_1d('matter_power_lin', 'z')

    # If hmf and input k_vec to match!
    # config_hmf['lnk_min'] = np.log(k_vec_original[0])
    # config_hmf['lnk_max'] = np.log(k_vec_original[-1])
    # config_hmf['dlnk'] = (config_hmf['lnk_max'] - config_hmf['lnk_min']) / config_hmf['nk']

    plin = log_linear_interpolation_k(plin_original, k_vec_original, k_vec)

    power_kwargs = {
        'matter_power_lin': plin,
        'mead_correction': mead_correction,
        'dewiggle': dewiggle,
        'k_vec': k_vec,
        'z_vec': config_hmf['z_vec'],
        'lnk_min': config_hmf['lnk_min'],
        'lnk_max': config_hmf['lnk_max'],
        'dlnk': config_hmf['dlnk'],
        'Mmin': config_hmf['log_mass_min'],
        'Mmax': config_hmf['log_mass_max'],
        'dlog10m': config_hmf['dlog10m'],
        'mdef_model': config_hmf['mdef_model'],
        'hmf_model': config_hmf['hmf_model'],
        'bias_model': config_hmf['bias_model'],
        'halo_profile_model_dm': config_hmf['profile_cen'],
        'halo_concentration_model_dm': config_hmf['cm_model_cen'],
        'halo_profile_model_sat': config_hmf['profile_sat'],
        'halo_concentration_model_sat': config_hmf['cm_model_sat'],
        'transfer_model': 'FromArray',
        'transfer_params': {'k': transfer_k, 'T': transfer_func},
        'growth_model': 'FromArray',
        'growth_params': {'z': growth_z, 'd': growth_func},
        'norm_cen': block[config_hmf['profile_value_name'], 'norm_cen'],
        'norm_sat': block[config_hmf['profile_value_name'], 'norm_sat'],
        'eta_cen': block[config_hmf['profile_value_name'], 'eta_cen'],
        'eta_sat': block[config_hmf['profile_value_name'], 'eta_sat'],
        'overdensity': config_hmf['overdensity'],
        'delta_c': config_hmf['delta_c'],
        'one_halo_ktrunc': one_halo_ktrunc,
        'two_halo_ktrunc': two_halo_ktrunc,
        'omega_c': block[cosmo_params, 'omega_c'],
        'omega_b': block[cosmo_params, 'omega_b'],
        'h0': block[cosmo_params, 'h0'],
        'n_s': block[cosmo_params, 'n_s'],
        'sigma_8': block[cosmo_params, 'sigma_8'],
        'm_nu': block[cosmo_params, 'mnu'],
        'w0': block[cosmo_params, 'w'],
        'wa': block[cosmo_params, 'wa'],
        'tcmb': block.get_double(cosmo_params, 'TCMB', default=2.7255),
        'log10T_AGN': block['halo_model_parameters', 'logT_AGN'],
        'mb': 10.0 ** block['halo_model_parameters', 'm_b'],
    }

    if bnl:
        num_calls = cached_bnl['num_calls']
        update_bnl = cached_bnl['update_bnl']

        if num_calls % update_bnl == 0:
            bnl = NonLinearBias(
                mass=10
                ** np.arange(
                    config_hmf['log_mass_min'],
                    config_hmf['log_mass_max'],
                    config_hmf['dlog10m'],
                ),
                z_vec=z_vec,
                k_vec=k_vec,
                h0=block[cosmo_params, 'h0'],
                A_s=block[cosmo_params, 'A_s'],
                omega_b=block[cosmo_params, 'omega_b'],
                omega_c=block[cosmo_params, 'omega_c'],
                omega_lambda=1.0 - block[cosmo_params, 'omega_m'],
                n_s=block[cosmo_params, 'n_s'],
                w0=block[cosmo_params, 'w'],
                z_dep=False,
            )

            beta_interp = bnl.bnl
            cached_bnl['cached_bnl'] = beta_interp
        else:
            beta_interp = cached_bnl['cached_bnl']

        cached_bnl['num_calls'] = num_calls + 1

        power_kwargs.update(
            {
                'bnl': bnl,
                'beta_nl': beta_interp,
            }
        )

    if response or fortuna:
        k_nl, p_nl = get_nonlinear_power_spectrum(block, z_vec)
        pk_mm_in = log_linear_interpolation_k(p_nl, k_nl, k_vec)
        power_kwargs.update({'matter_power_nl': pk_mm_in, 'response': response})
    else:
        pk_mm_in = None

    if galaxy or alignment:
        poisson_params = {}
        for param in poisson_parameters[poisson_type]:
            if not block.has_value('pk_parameters', param):
                raise Exception(
                    f'Error: parameter {param} is needed for the requested poisson model: {poisson_type}'
                )
            poisson_params[param] = get_string_or_none(
                block, 'pk_parameters', param, default=None
            )

        hod_params['A_cen'] = (
            block[hod_values_name, 'A_cen']
            if block.has_value(hod_values_name, 'A_cen')
            else None
        )
        hod_params['A_sat'] = (
            block[hod_values_name, 'A_sat']
            if block.has_value(hod_values_name, 'A_sat')
            else None
        )
        hod_parameters = parameters_models[hod_model]
        # Dinamically load required HOD parameters givent the model and number of bins!
        for param in hod_parameters:
            if hod_model == 'Cacciato':
                param_bin = param
                if not block.has_value(hod_values_name, param_bin):
                    raise Exception(
                        f'Error: parameter {param} is needed for the requested hod model: {hod_model}'
                    )
                hod_params[param] = block[hod_values_name, param_bin]
            else:
                param_list = []
                for nb in range(nbins):
                    suffix = f'_{nb + 1}' if nbins != 1 else ''
                    param_bin = f'{param}{suffix}'
                    if not block.has_value(hod_values_name, param_bin):
                        raise Exception(
                            f'Error: parameter {param} is needed for the requested hod model: {hod_model}'
                        )
                    param_list.append(block[hod_values_name, param_bin])
                hod_params[param] = np.array(param_list)

        power_kwargs.update(
            {
                'poisson_model': poisson_type,
                'poisson_params': poisson_params,
                'pointmass': point_mass,
                'hod_model': hod_model,
                'hod_params': hod_params,
                'hod_settings': hod_settings,
                'obs_settings': obs_settings,
                'compute_observable': obs_settings['save_observable'],
                'hod_settings_mm': hod_settings_mm,
            }
        )

    # TO-DO: for aligments at least we need to split the calculation in red/blue and add here!

    if alignment:
        align_params = {}
        if central_IA == 'halo_mass':
            align_params.update(
                {
                    'beta_sat': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_1}', 'beta_sat'
                    ],
                    'mpivot_sat': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_1}', 'M_pivot'
                    ],
                }
            )
        else:
            align_params.update(
                {
                    'beta_sat': None,
                    'mpivot_sat': None,
                }
            )
        if satellite_IA == 'halo_mass':
            align_params.update(
                {
                    'beta_cen': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_1}', 'beta'
                    ],
                    'mpivot_cen': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_1}', 'M_pivot'
                    ],
                }
            )
        else:
            align_params.update(
                {
                    'beta_cen': None,
                    'mpivot_cen': None,
                }
            )
        align_params.update(
            {
                'nmass': 5,
                'n_hankel': 350,
                'nk': 10,
                'ell_max': 6,
                'gamma_1h_slope': block[
                    f'intrinsic_alignment_parameters{pop_name_ia_1}',
                    'gamma_1h_radial_slope',
                ],
                'gamma_1h_amplitude': block[
                    f'intrinsic_alignment_parameters{pop_name_ia_1}',
                    'gamma_1h_amplitude',
                ],
                'gamma_2h_amplitude': block[
                    f'intrinsic_alignment_parameters{pop_name_ia_1}',
                    'gamma_2h_amplitude',
                ],
            }
        )
        power_kwargs.update(
            {
                'align_params': align_params,
                'one_halo_ktrunc_ia': one_halo_ktrunc_ia,
                'two_halo_ktrunc_ia': two_halo_ktrunc_ia,
                'fortuna': fortuna,
                't_eff': block.get_double(
                    'pk_parameters', 'linear_fraction_fortuna', default=0.0
                ),
            }
        )
        if split_ia:
            align_params_2 = align_params.copy()
            if central_IA == 'halo_mass':
                align_params_2.update(
                    {
                        'beta_sat': block[
                            f'intrinsic_alignment_parameters{pop_name_ia_2}', 'beta_sat'
                        ],
                        'mpivot_sat': block[
                            f'intrinsic_alignment_parameters{pop_name_ia_2}', 'M_pivot'
                        ],
                    }
                )
            else:
                align_params_2.update(
                    {
                        'beta_sat': None,
                        'mpivot_sat': None,
                    }
                )
            if satellite_IA == 'halo_mass':
                align_params_2.update(
                    {
                        'beta_cen': block[
                            f'intrinsic_alignment_parameters{pop_name_ia_2}', 'beta'
                        ],
                        'mpivot_cen': block[
                            f'intrinsic_alignment_parameters{pop_name_ia_2}', 'M_pivot'
                        ],
                    }
                )
            else:
                align_params_2.update(
                    {
                        'beta_cen': None,
                        'mpivot_cen': None,
                    }
                )
            align_params_2.update(
                {
                    'gamma_1h_slope': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_2}',
                        'gamma_1h_radial_slope',
                    ],
                    'gamma_1h_amplitude': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_2}',
                        'gamma_1h_amplitude',
                    ],
                    'gamma_2h_amplitude': block[
                        f'intrinsic_alignment_parameters{pop_name_ia_2}',
                        'gamma_2h_amplitude',
                    ],
                }
            )

    hod = not (matter and not galaxy)
    # power = Spectra(**power_kwargs)
    power = config_hmf['power']
    power.update(**power_kwargs)

    results = UpsampledSpectra(
        z=z_vec_original,
        k=k_vec_original,
        fraction_z=None,
        fraction=None,
        model=power,
        model_1_params=power_kwargs,
        model_2_params=None,
    )

    if matter:
        save_matter_results(block, power, z_vec, k_vec)

    if hod:
        hod_bins = save_hod_results(block, power, z_vec, hod_section_name, hod_settings)
        if power.obs_func is not None and obs_settings['save_observable']:
            save_obs_results(
                block, power, obs_settings['observable_section_name'], obs_settings
            )

    z_out = z_vec_original
    k_out = k_vec_original
    if p_mm or response:
        results.results(['mm'], ['1h', '2h', 'tot'])
        pk_mm_1h = results.power_spectrum_mm.pk_1h
        pk_mm_2h = results.power_spectrum_mm.pk_2h
        pk_mm = results.power_spectrum_mm.pk_tot
        if response:
            # Here we save the computed Pmm to datablock as matter_power_hm,
            # but not replacing the Pnl with it, as in the response
            # method, the Pnl stays the same as one from CAMB
            block.put_grid(
                'matter_power_hm', 'z', z_out, 'k_h', k_out, 'p_k_1h', pk_mm_1h[0]
            )
            block.put_grid(
                'matter_power_hm', 'z', z_out, 'k_h', k_out, 'p_k_2h', pk_mm_2h[0]
            )
            block.put_grid('matter_power_hm', 'z', z_out, 'k_h', k_out, 'p_k', pk_mm[0])
        else:
            block.put_grid(
                'matter_power_nl', 'z', z_out, 'k_h', k_out, 'p_k_1h', pk_mm_1h[0]
            )
            block.put_grid(
                'matter_power_nl', 'z', z_out, 'k_h', k_out, 'p_k_2h', pk_mm_2h[0]
            )
            block.put_grid('matter_power_nl', 'z', z_out, 'k_h', k_out, 'p_k', pk_mm[0])
    else:
        pk_mm = None

    if p_gg:
        results.results(['gg'], ['1h', '2h', 'tot'])
        pk_gg_1h = results.power_spectrum_gg.pk_1h
        pk_gg_2h = results.power_spectrum_gg.pk_2h
        pk_gg = results.power_spectrum_gg.pk_tot
        bg_linear = power.power_spectrum_gg.galaxy_linear_bias
        for nb in range(hod_bins):
            suffix = f'_{nb + 1}' if hod_bins != 1 else ''
            save_pk_to_grid(
                block,
                z_out,
                k_out,
                'galaxy_power',
                suffix,
                pk_gg_1h[nb],
                pk_gg_2h[nb],
                pk_gg[nb],
            )
            block.put_grid(
                f'galaxy_linear_bias{suffix}',
                'z',
                z_vec,
                'k_h',
                k_vec,
                'bg_linear',
                bg_linear[nb],
            )

    if p_gm:
        results.results(['gm'], ['1h', '2h', 'tot'])
        pk_gm_1h = results.power_spectrum_gm.pk_1h
        pk_gm_2h = results.power_spectrum_gm.pk_2h
        pk_gm = results.power_spectrum_gm.pk_tot
        bgm_linear = power.power_spectrum_gm.galaxy_linear_bias
        for nb in range(hod_bins):
            suffix = f'_{nb + 1}' if hod_bins != 1 else ''
            save_pk_to_grid(
                block,
                z_out,
                k_out,
                'matter_galaxy_power',
                suffix,
                pk_gm_1h[nb],
                pk_gm_2h[nb],
                pk_gm[nb],
            )
            block.put_grid(
                f'galaxy_matter_linear_bias{suffix}',
                'z',
                z_vec,
                'k_h',
                k_vec,
                'bgm_linear',
                bgm_linear[nb],
            )

    if alignment:
        power.update(hod_settings=hod_settings_ia_1)
        hod_bins = save_hod_results(
            block, power, z_vec, hod_section_name_ia_1, hod_settings_ia_1
        )

        if split_ia:
            model_2_params = {
                'hod_settings': hod_settings_ia_2,
                'align_params': align_params_2,
            }
            results = UpsampledSpectra(
                z=z_vec_original,
                k=k_vec_original,
                fraction_z=z_vec_original,
                fraction=np.ones_like(z_vec_original),
                model=power,
                model_1_params=power_kwargs,
                model_2_params=model_2_params,
            )

        if p_II:
            results.results(['ii'], ['1h', '2h', 'tot'])
            pk_II_1h = results.power_spectrum_ii.pk_1h
            pk_II_2h = results.power_spectrum_ii.pk_2h
            pk_II = results.power_spectrum_ii.pk_tot
            for nb in range(hod_bins):
                suffix = f'_{nb + 1}' if hod_bins != 1 else ''
                save_pk_to_grid(
                    block,
                    z_out,
                    k_out,
                    'intrinsic_power',
                    suffix,
                    pk_II_1h[nb],
                    pk_II_2h[nb],
                    pk_II[nb],
                )

        if p_gI:
            results.results(['gi'], ['1h', '2h', 'tot'])
            pk_gI_1h = results.power_spectrum_gi.pk_1h
            pk_gI_2h = results.power_spectrum_gi.pk_2h
            pk_gI = results.power_spectrum_gi.pk_tot
            for nb in range(hod_bins):
                suffix = f'_{nb + 1}' if hod_bins != 1 else ''
                save_pk_to_grid(
                    block,
                    z_out,
                    k_out,
                    'galaxy_intrinsic_power',
                    suffix,
                    pk_gI_1h[nb],
                    pk_gI_2h[nb],
                    pk_gI[nb],
                )

        if p_mI:
            results.results(['mi'], ['1h', '2h', 'tot'])
            pk_mI_1h = results.power_spectrum_mi.pk_1h
            pk_mI_2h = results.power_spectrum_mi.pk_2h
            pk_mI = results.power_spectrum_mi.pk_tot
            for nb in range(hod_bins):
                suffix = f'_{nb + 1}' if hod_bins != 1 else ''
                save_pk_to_grid(
                    block,
                    z_out,
                    k_out,
                    'matter_intrinsic_power',
                    suffix,
                    pk_mI_1h[nb],
                    pk_mI_2h[nb],
                    pk_mI[nb],
                )

    config_hmf['power'] = power
    return 0


def cleanup(config):
    # Usually python modules do not need to do anything here.
    # We just leave it in out of pedantic completeness.
    pass



import astropy.units
import copy
import numpy as np
import os
import warnings
from astropy.io import fits
from astropy.table import Table
from enum import Enum

# FITS header keyword indicating 2-point data
TWOPOINT_SENTINEL = "2PTDATA"
NZ_SENTINEL = "NZDATA"
COV_SENTINEL = "COVDATA"
window_types = ["SAMPLE", "CLBP", "LOG_MID"]
# LOG_MID interpolates at log mid of bin

METADATA_PREFIX = "MD_"

# Please do not add things to this list
ANGULAR_UNIT_TYPES = [
    astropy.units.arcsec,
    astropy.units.arcmin,
    astropy.units.rad,
    astropy.units.deg,
]
ANGULAR_UNITS = {unit.name: unit for unit in ANGULAR_UNIT_TYPES}


# FITS header keyword indicating 1-point data
ONEPOINT_SENTINEL = "1PTDATA"

# Please do not add things to this list
#ONEPOINT_UNIT_TYPES = [
#    astropy.units.Lsun,
#    astropy.units.Msun,
#    astropy.units.mag,
#]
#ONEPOINT_UNITS = {unit.name: unit for unit in ONEPOINT_UNIT_TYPES}


def sample_cov(xi_arrays, mode='full'):
    """mode should be full, subsample or jk"""
    nsample, npoints = xi_arrays.shape
    Cov = np.zeros((npoints, npoints))
    Corr = np.zeros_like(Cov)

    xi_mean = np.mean(xi_arrays, axis=0)
    for i in range(npoints):
        for j in range(npoints):
            Cov[i, j] = np.sum((xi_arrays[:, i] - xi_mean[i])
                               * (xi_arrays[:, j] - xi_mean[j])) / nsample
    # This is the covariance in a patch of size A/nsample. Assume Cov ~ 1/A so Cov=Cov/nsample
    if mode == 'subsample':
        Cov /= nsample
    elif mode == "jk":
        Cov *= (nsample - 1)
    for i in range(npoints):
        for j in range(npoints):
            Corr[i, j] = Cov[i, j] / np.sqrt(Cov[i, i] * Cov[j, j])
    return Cov, Corr


class Types(Enum):
    """
    This is an enumeration - a list of possible values with names and code values
    that we can use in FITS headers.

    It enumerates the different quantities that a two-point measurement can correlate.
    For example, CMB T,E,B, lensing E and B, galaxy position and magnification.

    It specifies the quantity and whether we are in Fourier space or real space.

    One special case is xi_{-} and xi_{+} in galaxy shear.  These values are already
    correlations of combinations of E and B modes. We denote this as xi_{++} and xi_{--} here.

    """
    galaxy_position_fourier = "GPF"
    galaxy_shear_emode_fourier = "GEF"
    galaxy_shear_bmode_fourier = "GBF"
    galaxy_position_real = "GPR"
    galaxy_shear_plus_real = "G+R"
    galaxy_shear_minus_real = "G-R"
    cmb_kappa_real = "CKR"
    cmb_kappa_fourier = "CKF"
    onepoint = "O"

    @classmethod
    def lookup(cls, value):
        for T in cls:
            if T.value == value:
                return T


def dummy_kernel(name):
    return NumberDensity(name, np.zeros(10), np.zeros(10), np.zeros(10), [np.zeros(10)])


class OnePointMeasurement:
    """
    This class contains 1-pt information for a collection of numbered observable bins.
    It is expected to be used for a single 1-pt type (e.g. stellar mass function) split into
    stellar mass/luminosity bins.

    Since the main use case for this is 1-pt function, and measurement codes typically
    produce 2-pt like data sets, but one can pass a histogram as well.
    """

    def __init__(self, name, obs, nobs, obs_low=None, obs_high=None):
        self.name = name
        self.obs = obs
        self.nbin = len(nobs)
        self.obs_low = obs_low
        self.obs_high = obs_high
        self.nsample = []
        for i in range(self.nbin):
            self.nsample.append(len(nobs[i]))
        self.nobs = nobs

    def __str__(self):
        return f"<One Point: {self.name}>"

    def __repr__(self):
        return f"<One Point: {self.name}>"

    @classmethod
    def from_fits(cls, extension):
        # load in the 1-pt data
        data = extension.data
        header = extension.header

        i = 1
        name = f'VALUE{i}'
        name_x = f'ANG{i}'
        name_low = f'ANGLEMIN{i}'
        name_high = f'ANGGLEMAX{i}'
        nobs = []
        obs = []
        obs_low = []
        obs_high = []

        while name in data.names:
            obs_in = data[name]
            nobs.append(obs_in)
            obs.append(data[f'ANG{i}'])
            if f'ANGLEMIN{i}' in data.names:
                obs_low.append(data[f'ANGLEMIN{i}'])
            if f'ANGLEMAX{i}' in data.names:
                obs_high.append(data[f'ANGLEMAX{i}'])
            i += 1
            name = f'VALUE{i}'

        if obs_low == []:
            obs_low = None
        if obs_high == []:
            obs_high = None

        N = cls(extension.name, obs, nobs, obs_low, obs_high)

        return N

    def to_fits(self):
        header = fits.Header()
        header[ONEPOINT_SENTINEL] = True
        header['EXTNAME'] = self.name
        header['NBINS'] = self.nbin
        header['QUANT1'] = "O"
        header['QUANT2'] = "O"

        columns = []

        for i in range(self.nbin):
            name = f'VALUE{i + 1}'
            columns.append(fits.Column(name=name, array=self.nobs[i], format='D'))
            columns.append(fits.Column(name=f'ANG{i + 1}', array=self.obs[i], format='D'))
            if self.obs_low is not None:
                columns.append(fits.Column(name=f'ANGLEMIN{i + 1}', array=self.obs_low[i], format='D'))
            if self.obs_high is not None:
                columns.append(fits.Column(name=f'ANGLEMAX{i + 1}', array=self.obs_high[i], format='D'))

        extension = fits.BinTableHDU.from_columns(columns, header=header)
        return extension

    @classmethod
    def from_block(cls, block, section_name, output_name=None):
        """Load 1pt from cosmosis datablock"""
        if output_name == None:
            output_name = section_name

        nobs = []
        obs = []
        obs_low = []
        obs_high = []
        for i in range(1, 99999):
            bin_name = "bin_%d" % i
            if block.has_value(section_name, bin_name):
                nobs.append(block[section_name, bin_name])
            else:
                break
            if block.has_value(section_name, f"obs_low_{i}"):
                obs.append(block[section_name, f"obs_{i}"])
                obs_low.append(block[section_name, f"obs_low_{i}"])
                obs_high.append(block[section_name, f"obs_high_{i}"])
            else:
                obs_in = block[section_name, f"obs_{i}"]
                obs.append(obs_in)
                #d_obs = np.log10(obs_in[2]) - np.log10(obs_in[1])
                #obs_low.append(10.0**(np.log10(obs_in) - 0.5 * d_obs))
                #obs_high.append(10.0**(np.log10(obs_in) + 0.5 * d_obs))
        if obs_low == []:
            obs_low = None
        if obs_high == []:
            obs_high = None

        N = cls(output_name, obs, nobs, obs_low, obs_high)
        return N



class NumberDensity:
    """
    This class contains n(z) information for a collection of numbered redshift bins.
    It is expected to be used for a single sample type (e.g. galaxy sample) split into
    tomographic bins rather than the more complex cases where there are multiple different
    quantities (e.g. two galaxy samples).

    Since the main use case for this is photometric redshifts, and photo-z codes typically
    produce histogram type data sets (that is, they look like step functions between each bin),
    that is what this form assumes.
    """

    def __init__(self, name, zlow, z, zhigh, nzs, ngal=None, sigma_e=None):
        self.name = name
        self.zlow = zlow
        self.z = z
        self.zhigh = zhigh
        self.nbin = len(nzs)
        if self.nbin > 0:
            self.nsample = len(nzs[0])
        else:
            self.nsample = 0
        self.nzs = nzs
        self.ngal = ngal
        self.sigma_e = sigma_e

    @classmethod
    def from_fits(cls, extension):
        # load in the n(z) data
        data = extension.data
        header = extension.header

        z = data['Z_MID']
        zlow = data['Z_LOW']
        zhigh = data['Z_HIGH']
        i = 1
        name = f'BIN{i}'
        nzs = []
        sigma_e = []
        ngal = []
        while name in data.names:
            nz = data[name]
            nzs.append(nz)
            ngal.append(header.get(f'NGAL_{i}'))
            sigma_e.append(header.get(f'SIG_E_{i}'))
            i += 1
            name = f'BIN{i}'

        if all(x is None for x in sigma_e):
            sigma_e = None
        else:
            assert not any(x is None for x in sigma_e), "Please specify all or none of the SIG_E_ in your n(z) section {}".format(
                extension.name)
            sigma_e = np.array(sigma_e)

        if all(x is None for x in ngal):
            ngal = None
        else:
            assert not any(x is None for x in ngal), "Please specify all or none of the NGAL in your n(z) section {}".format(
                extension.name)
            ngal = np.array(ngal)

        N = cls(extension.name, zlow, z, zhigh,
                nzs, ngal=ngal, sigma_e=sigma_e)

        return N

    def to_fits(self):
        header = fits.Header()
        header[NZ_SENTINEL] = True
        header['EXTNAME'] = self.name

        columns = [
            fits.Column(name='Z_LOW', array=self.zlow, format='D'),
            fits.Column(name='Z_MID', array=self.z, format='D'),
            fits.Column(name='Z_HIGH', array=self.zhigh, format='D'),
        ]

        for i in range(self.nbin):
            name = f"BIN{i + 1}"
            columns.append(fits.Column(
                name=name, array=self.nzs[i], format='D'))

        if self.sigma_e is not None:
            for i in range(self.nbin):
                name = f"SIG_E_{i + 1}"
                header[name] = self.sigma_e[i]

        if self.ngal is not None:
            for i in range(self.nbin):
                name = f"NGAL_{i + 1}"
                header[name] = self.ngal[i]

        extension = fits.BinTableHDU.from_columns(columns, header=header)
        return extension

    @classmethod
    def from_block(cls, block, section_name, output_name=None):
        """Load kernel from cosmosis datablock"""
        if output_name == None:
            output_name = section_name
        if block.has_value(section_name, "z_mid"):
            z_mid = block[section_name, "z_mid"]
            z_low = block[section_name, "z_low"]
            z_high = block[section_name, "z_high"]
        else:
            z_mid = block[section_name, "z"]
            dz = z_mid[2] - z_mid[1]
            z_low = z_mid - 0.5 * dz
            z_high = z_mid + 0.5 * dz

        nzs = []
        for i in range(1, 99999):
            bin_name = "bin_%d" % i
            if block.has_value(section_name, bin_name):
                nzs.append(block[section_name, bin_name])
            else:
                break

        N = cls(output_name, z_low, z_mid, z_high, nzs)
        return N


class SpectrumMeasurement:
    def __init__(self, name, bins, types, kernels, windows, angular_bin, value,
                 angle=None, error=None, angle_unit=None, metadata=None, npairs=None,
                 varxi=None, extra_cols=None, angle_min=None, angle_max=None):
        """
        Metadata is a dictionary which will get added to the fits header
        extra cols is a dictionary of tuples
        """
        self.name = name
        self.bin1, self.bin2 = bins
        self.bin_pairs = self.get_bin_pairs()  # unique bin pairs
        self.type1, self.type2 = types
        self.kernel1, self.kernel2 = kernels
        self.angular_bin = angular_bin
        self.angle = angle
        self.angle_min = angle_min
        self.angle_max = angle_max
        self.value = value
        self.npairs = npairs
        self.varxi = varxi
        if windows in window_types:
            self.windows = windows
        else:
            raise TypeError("window type %s not recognised" % windows)
        self.error = error
        self.metadata = metadata
        if self.is_real_space():
            #angle is real
            msg = "Files with real-space units must specify units as one of: {}".format(
                list(ANGULAR_UNITS.keys()))
            assert angle_unit in ANGULAR_UNITS,  msg
        self.angle_unit = angle_unit
        # This is the index in the datavector, will be populated when used in TwoPointFile
        self.dv_index = None
        self.extra_cols = extra_cols

    def __str__(self):
        return f"<Spectrum: {self.name}>"

    def __repr__(self):
        return f"<Spectrum: {self.name}>"

    def get_bin_pairs(self):
        unique_pairs = []
        for p in zip(self.bin1, self.bin2):
            if p not in unique_pairs:
                unique_pairs.append(p)
        return unique_pairs

    def canonical_order(self):
        # order by b1, then b2, then angle
        order = np.lexsort([self.bin1, self.bin2, self.angular_bin])
        return order

    def is_real_space(self):
        return self.type1.value.endswith("R") or self.type2.value.endswith("R")

    def recompute_angular_bins(self):
        angles = np.unique(self.angle)
        angles.sort()
        angles = angles.tolist()
        angular_bin = [angles.index(ang)+1 for ang in self.angle]
        self.angular_bin = np.array(angular_bin)

    def convert_angular_units(self, unit):
        if not self.is_real_space():
            raise ValueError(
                "Two point spectrum has no units to convert; it is in Fourier space")

        if self.windows not in ["SAMPLE", "CLBP"]:
            raise NotImplementedError(
                "Need to write code to transform window function units")
        old_unit = ANGULAR_UNITS[self.angle_unit]
        new_unit = ANGULAR_UNITS[unit]

        angle_with_units = self.angle * old_unit
        self.angle = angle_with_units.to(new_unit).value
        self.angle_unit = unit
        if self.angle_min is not None:
            self.angle_min = (self.angle_min * old_unit).to(new_unit).value
            self.angle_max = (self.angle_max * old_unit).to(new_unit).value

    def apply_mask(self, mask):
        """mask is a boolean array which True for elements to be kept"""
        self.bin1 = self.bin1[mask]
        self.bin2 = self.bin2[mask]
        self.angular_bin = self.angular_bin[mask]
        self.angle = self.angle[mask]
        self.value = self.value[mask]
        if self.error is not None:
            self.error = self.error[mask]
        if self.angle_min is not None:
            self.angle_min = self.angle_min[mask]
            self.angle_max = self.angle_max[mask]

    def cut_bin_pair(self, bin_pair, complain=False):
        """Cut a full bin pair. If complain is True,
        raise a ValueError if the bin pair is not found"""
        b1, b2 = bin_pair
        mask = (self.bin1 == b1) * (self.bin2 == b2)
        if (mask.sum() == 0) and complain:
            raise ValueError("""You tried to cut bin pair %d,%d from spectrum named %s
                             but it doesn't exist""" % (b1, b2, self.name))
        elif mask.sum() > 0:
            self.apply_mask(~mask)

    def auto_bins(self):
        return self.bin1 == self.bin2

    def __len__(self):
        return len(self.value)

    def nbin(self):
        return np.max([self.bin1.max(), self.bin2.max()])

    def get_pair(self, bin1, bin2):
        w = (self.bin1 == bin1) & (self.bin2 == bin2)
        return self.angle[w], self.value[w]

    def get_pair_mask(self, bin1, bin2):
        w = (self.bin1 == bin1) & (self.bin2 == bin2)
        return w

    def get_error(self, bin1, bin2):
        if self.error is None:
            return None
        w = (self.bin1 == bin1) & (self.bin2 == bin2)
        return self.error[w]

    @classmethod
    def from_fits(cls, extension, covmat_info=None):
        name = extension.name
        # determine the type of the quantity involved
        type1 = Types.lookup(extension.header['QUANT1'])
        type2 = Types.lookup(extension.header['QUANT2'])

        # and the name of the kernels and the window function
        # extensions
        kernel1 = extension.header['KERNEL_1']
        kernel2 = extension.header['KERNEL_2']
        windows = extension.header['WINDOWS']

        if windows not in window_types:
            raise NotImplementedError(
                "Have not yet coded window functions for angular bins")

        # Now load the data
        data = extension.data
        bin1 = data['BIN1']
        bin2 = data['BIN2']
        angular_bin = data['ANGBIN']
        value = data['VALUE']
        if "ANG" in data.names:
            angle = data['ANG']
            ang_index = data.names.index("ANG")
            angle_unit = extension.header.get(f'TUNIT{ang_index + 1}')
            if angle_unit is not None:
                angle_unit = angle_unit.strip()
        else:
            angle = None
            angle_unit = None
        if "NPAIRS" in data.names:
            npairs = data['NPAIRS']
        else:
            npairs = None
        if "VARXI" in data.names:
            varxi = data['VARXI']
        else:
            varxi = None
        if "ANGLEMIN" in data.names:
            angle_min = data['ANGLEMIN']
        else:
            angle_min = None
        if "ANGLEMAX" in data.names:
            angle_max = data['ANGLEMAX']
        else:
            angle_max = None

        # check for extra columns
        found_extra_cols = False
        for c in data.names:
            if c.startswith("XTRA_"):
                if not found_extra_cols:
                    extra_cols = {}
                    found_extra_cols = True
                colname = (c.replace("XTRA_", "")).lower()
                extra_cols[colname] = data[c]
        if not found_extra_cols:
            extra_cols = None

        # check for metadata
        metadata = {}
        for key in extension.header:
            if key.startswith(METADATA_PREFIX):
                metadata[key.replace(METADATA_PREFIX, "")
                         ] = extension.header[key]

        # Load a chunk of the covariance matrix too if present.
        if covmat_info is None:
            error = None
        else:
            error = covmat_info.get_error(name)

        return SpectrumMeasurement(name, (bin1, bin2), (type1, type2), (kernel1, kernel2), windows,
                                   angular_bin, value, angle, error, angle_unit=angle_unit, npairs=npairs,
                                   varxi=varxi, angle_min=angle_min, angle_max=angle_max,
                                   extra_cols=extra_cols, metadata=metadata)

    def to_fits(self):
        header = fits.Header()
        header[TWOPOINT_SENTINEL] = True
        header['EXTNAME'] = self.name
        header['QUANT1'] = self.type1.value
        header['QUANT2'] = self.type2.value
        header['KERNEL_1'] = self.kernel1
        header['KERNEL_2'] = self.kernel2
        header['WINDOWS'] = self.windows  # NOT YET CODED ANYTHING ELSE
        header['N_ZBIN_1'] = len(np.unique(self.bin1))
        header['N_ZBIN_2'] = len(np.unique(self.bin2))
        if (self.metadata is not None) and self.metadata:
            # Check metadata doesn't share any keys with the stuff that's already in the header
            assert set(self.metadata.keys()).isdisjoint(set(header.keys()))
            for key, val in self.metadata.iteritems():
                # Use MD_ prefix so metadata entries can be easily recognised by from_fits
                header[METADATA_PREFIX+key] = val
        header['N_ANG'] = len(np.unique(self.angular_bin))

        columns = [
            fits.Column(name='BIN1', array=self.bin1, format='K'),
            fits.Column(name='BIN2', array=self.bin2, format='K'),
            fits.Column(name='ANGBIN', array=self.angular_bin, format='K'),
            fits.Column(name='VALUE', array=self.value, format='D'),
        ]

        if self.angle_min is not None:
            columns.append(fits.Column(
                    name='ANGLEMIN', array=self.angle_min, format='D', unit=self.angle_unit))
        if self.angle_max is not None:
            columns.append(fits.Column(
                    name='ANGLEMAX', array=self.angle_max, format='D', unit=self.angle_unit))

        if self.angle is not None:
            if self.windows == "SAMPLE":
                columns.append(fits.Column(
                    name='ANG', array=self.angle, format='D', unit=self.angle_unit))
            if self.windows == "CLBP":
                columns.append(fits.Column(
                    name='ANG', array=self.angle, format='2K', unit=self.angle_unit))
        if self.npairs is not None:
            columns.append(fits.Column(
                name='NPAIRS', array=self.npairs, format='D'))
        if self.varxi is not None:
            columns.append(fits.Column(
                name='VARXI', array=self.varxi, format='D'))
        if self.extra_cols is not None:
            for (colname, arr) in self.extra_cols.iteritems():
                columns.append(fits.Column(
                    name='XTRA_'+colname, array=arr, format='D'))

        extension = fits.BinTableHDU.from_columns(columns, header=header)
        return extension


class CovarianceMatrixInfo:
    """Encapsulate a covariance matrix and indices in it"""

    def __init__(self, name, names, lengths, covmat):
        super().__init__()
        self.name = name
        self.names = names
        self.lengths = lengths
        self.starts = [0]
        for i, l in enumerate(self.lengths[:-1]):
            self.starts.append(l + self.starts[i])
        self.covmat = covmat
        self.diagonal = covmat.diagonal()

    def get_error(self, name):
        i = self.names.index(name)
        start = self.starts[i]
        end = start + self.lengths[i]
        return self.diagonal[start:end]**0.5

    def to_fits(self):
        header = fits.Header()

        header[COV_SENTINEL] = True
        header['EXTNAME'] = self.name
        for i, (start_index, name) in enumerate(zip(self.starts, self.names)):
            header[f'STRT_{i}'] = start_index
            header[f'NAME_{i}'] = name
        extension = fits.ImageHDU(data=self.covmat, header=header)
        return extension

    @classmethod
    def from_fits(cls, extension):
        cov_name = extension.name
        covmat = extension.data
        header = extension.header
        i = 0
        measurement_names = []
        start_indices = []
        while True:
            name_card = f'NAME_{i}'
            if name_card not in header:
                break
            measurement_names.append(header[name_card])
            start_index_card = f'STRT_{i}'
            start_indices.append(header[start_index_card])
            i += 1
        lengths = []
        current_length = 0
        # this only works if more than one spectrum
        if len(start_indices) > 1:
            for start, end in zip(start_indices[:-1], start_indices[1:]):
                lengths.append(end - start)
            if start_indices:
                lengths.append(covmat.shape[0] - start_indices[-1])
        else:
            lengths.append(covmat.shape[0])
        return cls(cov_name, measurement_names, lengths, covmat)

    @classmethod
    def from_spec_lists(cls, spec_lists, cov_name, mode='full'):
        """Often the covariance will be computed by measuring the statistic(s) in question
        on many simulated realisations of the dataset. This function takes a list of such
        measurements, *each one a list SpectrumMeasurement objects*, computes the mean and covariance, and
        returns the mean as a list of SpectrumMeasurements, and the covariance as a CovarianceMatrixInfo
        object. mode should be one of full, subsample or jackknife"""
        # first check that spec_lists is a list of lists, and that there are at least 2
        print('spec_lists', spec_lists)
        try:
            spec_lists[1][0]
        except Exception as e:
            print("spec_lists should be a list of lists with at least two elements")
            raise(e)

        # Get measurement names and lengths from first element of spec_lists
        num_spec = len(spec_lists[0])
        names = [s.name for s in spec_lists[0]]
        lengths = [len(s.value) for s in spec_lists[0]]

        # Now loop through all realisations, building up list of numpy arrays of raw measurements
        n_real = len(spec_lists)
        spec_arrays = []
        for i_real in range(n_real):
            spec_array = []
            # check this realisation has right number,type,length of spectra
            for i_spec in range(num_spec):
                assert spec_lists[i_real][i_spec].name == names[i_spec]
                assert len(spec_lists[i_real][i_spec].value) == lengths[i_spec]
                spec_array += list(spec_lists[i_real][i_spec].value)
            spec_arrays.append(np.array(spec_array))

        # Now compute covariance
        spec_arrays = np.array(spec_arrays)
        cov_values, _ = sample_cov(spec_arrays, mode=mode)
        mean_spec_values = np.mean(spec_arrays, axis=0)

        # make list of mean 2pt specs by copying spec_lists[0] and replacing value column
        mean_spec = copy.copy(spec_lists[0])
        index_start = 0
        for i_spec in range(num_spec):
            end = index_start + lengths[i_spec]
            inds = np.arange(index_start, index_start + lengths[i_spec])
            index_start = end
            mean_spec[i_spec].value = mean_spec_values[inds]

        return cls(cov_name, names, lengths, cov_values), mean_spec


class TwoPointFile:
    def __init__(self, spectra, kernels, nobs, windows, covmat_info):
        if windows is None:
            windows = {}
        self.spectra = spectra
        dv_start = 0
        for s in self.spectra:
            n_dv = len(s.value)
            s.dv_index = np.arange(dv_start, dv_start+n_dv)
            dv_start += n_dv
        self.kernels = kernels
        self.windows = windows
        self.nobs = nobs
        self.covmat_info = covmat_info
        if covmat_info:
            #self.covmat = covmat_info.covmat
            self.covmat = self.get_cov_start()
            for s in self.spectra:
                s.error = self.covmat_info.get_error(s.name)
        else:
            self.covmat = None
        self._spectrum_index = None

    def get_spectrum(self, name):
        spectra = [spectrum for spectrum in self.spectra if spectrum.name == name]
        n = len(spectra)
        if n == 0:
            raise ValueError("Spectrum with name %s not found in file" % name)
        elif n > 1:
            raise ValueError(
                "Multiple spectra with name %s found in file" % name)
        else:
            return spectra[0]

    def get_kernel(self, name):
        kernels = [kernel for kernel in self.kernels if kernel.name == name]
        n = len(kernels)
        if n == 0:
            raise ValueError("Kernel with name %s not found in file" % name)
        elif n > 1:
            raise ValueError(
                "Multiple kernel with name %s found in file" % name)
        else:
            return kernels[0]

    def get_nobs(self, name):
        nobs = [obs for obs in self.nobs if nobs.name == name]
        n = len(nobs)
        if n == 0:
            raise ValueError("Observable function with name %s not found in file" % name)
        elif n > 1:
            raise ValueError(
                "Multiple observable functions with name %s found in file" % name)
        else:
            return nobs[0]

    def _build_spectrum_index(self):
        index = 0
        self._spectrum_index = {}
        for spectrum in self.spectra:
            name = spectrum.name
            bin1 = spectrum.bin1
            bin2 = spectrum.bin2
            angbin = spectrum.angular_bin
            n = len(bin1)
            for i in range(n):
                self._spectrum_index[(
                    name, bin1[i], bin2[i], angbin[i])] = index
                index += 1

    def get_overall_index(self, spectrum_name, bin1, bin2, angbin):
        if self._spectrum_index is None:
            self._build_spectrum_index()
        return self._spectrum_index[(spectrum_name, bin1, bin2, angbin)]

    def _mask_covmat(self, masks):
        # Also cut down the covariance matrix accordingly
        if self.covmat is not None:
            mask = np.concatenate(masks)
            self.covmat = self.covmat[mask, :][:, mask]
            self.covmat_info = CovarianceMatrixInfo(self.covmat_info.name,
            [s.name for s in self.spectra] + [o.name for o in self.nobs], [len(s) for s in self.spectra] + [sum(o.nsample) for o in self.nobs], self.covmat)

    def mask_bad(self, bad_value):
        "Go through all the spectra masking out data where they are equal to bad_value"
        masks = []
        # go through the spectra and covmat, masking out the bad values.
        for spectrum in self.spectra:
            # nb this will not work for NaN!
            mask = (spectrum.value != bad_value)
            spectrum.apply_mask(mask)
            print("Masking {} values in {}".format(
                mask.size - mask.sum(), spectrum.name))
            # record the mask vector as we will need it to mask the covmat
            masks.append(mask)
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)
        if masks:
            self._mask_covmat(masks)

    def reorder_canonical(self):
        masks = []
        print("Reordering all data")
        n = 0
        for spectrum in self.spectra:
            mask = spectrum.canonical_order()
            spectrum.apply_mask(mask)
            masks.append(mask+n)
            n += len(mask)
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)
        if masks:
            self._mask_covmat(masks)

    def mask_indices(self, spectrum_name, indices):
        s = self.get_spectrum(spectrum_name)
        mask_points = np.array(indices)
        masks = []
        for s in self.spectra:
            mask = np.ones(len(s), dtype=bool)
            if s.name == spectrum_name:
                mask[mask_points] = False
                s.apply_mask(mask)
            print(f"Keeping {mask.sum()} points in {s.name}")
            masks.append(mask)
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)
        self._mask_covmat(masks)

    def mask_cross(self):
        masks = []
        for spectrum in self.spectra:
            mask = spectrum.auto_bins()
            spectrum.apply_mask(mask)
            print("Masking {} cross-values in {}".format(mask.size -
                                                         mask.sum(), spectrum.name))
            masks.append(mask)
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)
        if masks:
            self._mask_covmat(masks)

    def mask_scales(self, cuts={}, bin_cuts=[]):
        masks = []
        for spectrum in self.spectra:
            mask = np.ones(len(spectrum), dtype=bool)
            for b1, b2 in spectrum.bin_pairs:
                w_full = np.where((spectrum.bin1 == b1) &
                                  (spectrum.bin2 == b2))[0]
                if (spectrum.name, b1, b2) in bin_cuts:
                    #print("Removing {} bin ({},{}) altogether.".format(
                    #    spectrum.name, b1, b2))
                    mask[w_full] = False
                    continue

                cut = cuts.get((spectrum.name, b1, b2))
                if cut is None:
                    #print("No cut specified for {} bin ({},{})".format(
                    #    spectrum.name, b1, b2))
                    continue

                # Actually do the cut
                ang_min, ang_max = cut
                w = np.where((spectrum.bin1 == b1) & (spectrum.bin2 == b2) &
                             ((spectrum.angle < ang_min) | (spectrum.angle > ang_max)))[0]

                #print("Cutting {} bin pair ({},{}) to angle range ({} - {}):"
                #      " this removes {} values out of {}".format(
                #    spectrum.name, b1, b2, ang_min, ang_max, len(w), len(w_full)))

                mask[w] = False
            masks.append(mask)
            spectrum.apply_mask(mask)
            #print("")
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)

        if masks:
            self._mask_covmat(masks)

    def mask_scale(self, spectra_to_cut, min_scale=-np.inf, max_scale=np.inf):
        masks = []
        # go through the spectra and covmat, masking out the bad values.
        for spectrum in self.spectra:
            mask = np.ones(len(spectrum.value), dtype=bool)
            if (spectra_to_cut != "all") and (spectrum.name not in spectra_to_cut):
                masks.append(mask)
            else:
                # nb this will not work for NaN!
                mask = (spectrum.angle > min_scale) & (
                    spectrum.angle < max_scale)
                spectrum.apply_mask(mask)
                #print("Masking {} values in {} because they had ell or theta outside ({},{})".format(
                #    mask.size - mask.sum(), spectrum.name, min_scale, max_scale))
                # record the mask vector as we will need it to mask the covmat
                masks.append(mask)
        if self.nobs is not None:
            for obs in self.nobs:
                mask = np.ones(sum(obs.nsample), dtype=bool)
                masks.append(mask)
        if masks:
            self._mask_covmat(masks)

    def choose_data_sets(self, data_sets):
        """Strip out any data sets not in the given list."""
        data_sets = [d.lower() for d in data_sets]
        mask = []
        use = []
        use_obs = []
        for spectrum in self.spectra:
            if spectrum.name.lower() in data_sets:
                use.append(True)
                mask.append(np.ones(spectrum.bin1.size, dtype=bool))
            else:
                use.append(False)
                mask.append(np.zeros(spectrum.bin1.size, dtype=bool))
        if self.nobs is not None:
            for obs in self.nobs:
                if obs.name.lower() in data_sets:
                    use_obs.append(True)
                    mask.append(np.ones(sum(obs.nsample), dtype=bool))
                else:
                    use_obs.append(False)
                    mask.append(np.zeros(sum(obs.nsample), dtype=bool))
        for data_set in data_sets:
            if not any(spectrum.name.lower() == data_set for spectrum in self.spectra) and not any(obs.name.lower() == data_set for obs in self.nobs):
                raise ValueError(
                    f"Data set called {data_set} not found in two-point data file.")
        self.spectra = [s for (u, s) in zip(use, self.spectra) if u]
        if self.nobs is not None:
            self.nobs = [s for (u, s) in zip(use_obs, self.nobs) if u]
        if self.covmat is not None:
            mask = np.concatenate(mask)
            self.covmat = self.covmat[mask, :][:, mask]

    def get_cov_start(self):

        # This gets the covariance array in the right order (before any scale cuts etc.)
        cov = self.covmat_info.covmat
        if self.nobs is not None:
            names_list = [spec.name for spec in self.spectra] + [o.name for o in self.nobs]
        else:
            names_list = [spec.name for spec in self.spectra]
        if self.covmat_info.names == names_list:
            # Ordering is already ok
            return cov
        print("Covariance matrix is not in the same order as the 2pt measurement extensions...doing some damn fiddly")
        print("re-ordering, if I screw it up it's your fault for not putting your covariance in the right order")
        cov_starts = self.covmat_info.starts
        cov_lengths = self.covmat_info.lengths
        cov_names = self.covmat_info.names
        cov_ends = [cov_lengths[0]]
        for i in range(len(cov_names) - 1):
            cov_ends.append(cov_ends[i] + cov_lengths[i + 1])

        assert cov_ends[-1] == cov.shape[0]

        total_l = 0
        spec_inds = []
        if self.nobs is not None:
            spec_names = [spec.name for spec in self.spectra] + [o.name for o in self.nobs]
            specs = self.spectra + self.nobs
        else:
            spec_names = [spec.name for spec in self.spectra]
            specs = self.spectra
        for spectrum in specs:
            spec_inds.append(cov_names.index(spectrum.name))
            total_l += cov_lengths[cov_names.index(spectrum.name)]
        cov_out = np.zeros((total_l, total_l))
        start_i = 0

        for ti, ind_i in zip(spec_names, spec_inds):
            start_j = 0
            for tj, ind_j in zip(spec_names, spec_inds):
                cov_out[start_i:start_i + cov_lengths[ind_i], start_j:start_j + cov_lengths[ind_j]
                        ] = cov[cov_starts[ind_i]:cov_ends[ind_i], cov_starts[ind_j]:cov_ends[ind_j]]
                start_j += cov_lengths[ind_j]
            start_i += cov_lengths[ind_i]
        return cov_out

    def to_fits(self, filename, overwrite=False, clobber=False):
        hdus = [fits.PrimaryHDU()]

        if clobber:
            warnings.warn("The 'clobber' keyword in twopoint is deprecated."
                          "Please switch to overwrite.", DeprecationWarning)

        if self.covmat_info is not None:
            hdus.append(self.covmat_info.to_fits())

        for spectrum in self.spectra:
            if spectrum.windows not in window_types:
                raise NotImplementedError(
                    "Sorry - not yet coded general case with ell/theta window functions")
            hdus.append(spectrum.to_fits())

        if self.kernels is not None:
            for kernel in self.kernels:
                hdus.append(kernel.to_fits())

        if self.nobs is not None:
            for nobs in self.nobs:
                hdus.append(nobs.to_fits())

        hdulist = fits.HDUList(hdus)
        hdulist.writeto(filename, overwrite=(clobber or overwrite))

    @classmethod
    def from_fits(cls, filename, covmat_name="COVMAT"):
        fitsfile = fits.open(filename)
        spectra = []
        kernels = []
        nobs = []
        windows = {}

        # Load the covariance matrix from the file, typically stored as
        # image data.
        # It turns out it is more conventient to load this first.
        # We can also use no covmat at all.
        if covmat_name is None:
            covmat_info = None
            covmat = None
        else:
            extension = fitsfile[covmat_name]
            covmat_info = CovarianceMatrixInfo.from_fits(extension)

        # First load all the spectra in the file
        # Spectra are indicated by the "2PTDATA" keyword
        # in the header being "T"
        for extension in fitsfile:
            if extension.header.get(TWOPOINT_SENTINEL):
                spectra.append(SpectrumMeasurement.from_fits(
                    extension, covmat_info))

        # Each spectrum needs kernels, usually some n(z).
        # These were read from headers when we loaded the 2pt data above above.
        # Now we are loading those kernels into a dictionary
        for extension in fitsfile:
            if extension.header.get(NZ_SENTINEL):
                kernels.append(NumberDensity.from_fits(extension))

        # We might also have some 1pt functions saved.
        for extension in fitsfile:
            if extension.header.get(ONEPOINT_SENTINEL):
                nobs.append(OnePointMeasurement.from_fits(extension))

        # We might also require window functions W(ell) or W(theta). It's also possible
        # that we just use a single sample value of ell or theta instead.
        # If the spectra required it (according to the header) then we also
        # need to load those in.
        for spectrum in spectra:
            if spectrum.windows not in window_types and spectrum.windows not in windows:
                windows[spectrum.windows] = cls._windows_from_fits(
                    fitsfile[windows])

        # return a new TwoPointFile object with the data we have just loaded
        return cls(spectra, kernels, nobs, windows, covmat_info)

    @classmethod
    def _windows_from_fits(cls, extension):
        raise NotImplementedError("non-sample window functions in ell/theta")

    def plots(self, root, colormap='viridis', savepdf=False, latex=True, plot_spectrum=True, plot_kernel=True, plot_1pt=True, plot_cov=True, cov_vmin=None, save_pickle=False, load_pickle=False, remove_pickle=True, label_legend='', blind_yaxis=False, callback=None):
        """
        Makes plot of each for your spectra, kernels and covariance. Allows you to compare the spectra of different files.
        Options:
        - root: Name of the output plots.
        - colormap: Colormap used for the plots.
        - savepdf: True if you want to save pdf too, besides png.
        - latex: True if want to save with latex font. It will be slower. Set to false to test plot.
        - plot_spectrum, plot_kernel, plot_cov: whether or not to make this plots
        - cov_vmin = minimum value for the colorbar in the covariance plot.
        - plot_spectrum, plot_kernel and plot_cov are boolean variables that are true if you want to make these plots.
        - save_pickle: if true it saves a pickle file to edit be able to compare different files.
        - load_pickle: if true, it will continue the plot starting from a pickle file.
        - remove_pickle: set to true if you want to keep this file to edit your plot afterwards.
        - label_legend: name that will appear in the legend when comparing different files.
        - blind_axis: True if you want to remove the y-axis labels.
        """

        import matplotlib.pyplot as plt
        import pickle as pl
        from matplotlib import ticker

        if latex:
            plt.rc('text', usetex=True)
            plt.rc('font', family='serif')

        def savefig(name):
            print(f"Saving {name}")
            plt.savefig(name, bbox_inches='tight', dpi=400)
            if savepdf:
                plt.savefig(name + '.pdf', bbox_inches='tight')

        def corr_names(spectrum):
            '''
            Get latex labels and colors for each kind of correlation.
            '''
            if (spectrum.type1 == spectrum.type2 == Types.galaxy_shear_plus_real):
                corr_type = 'xip'
                label = r"$\xi_+(\theta)$"
                color = plt.get_cmap(colormap)(0)

            elif (spectrum.type1 == spectrum.type2 == Types.galaxy_shear_minus_real):
                corr_type = 'xim'
                label = r"$\xi_-(\theta)$"
                color = plt.get_cmap(colormap)(0.2)

            elif (spectrum.type1 == Types.galaxy_position_real) & (spectrum.type2 == Types.galaxy_shear_plus_real):
                corr_type = 'gt'
                label = r"$\gamma_t(\theta)$"
                color = plt.get_cmap(colormap)(0.4)

            elif (spectrum.type1 == spectrum.type2 == Types.galaxy_position_real):
                corr_type = 'wtheta'
                label = r"$w(\theta)$"
                color = plt.get_cmap(colormap)(0.6)

            else:
                corr_type = None
                label = None
                color = None

            return corr_type, label, color

        if plot_spectrum:

            for spectrum in self.spectra:

                corr_type, label, color = corr_names(spectrum)

                if corr_type is None:
                    continue

                mtype = 'o'
                # Use one color for each file in case you want to compare different files
                if save_pickle:
                    color = plt.get_cmap(colormap)(0)
                if save_pickle and load_pickle:
                    color = plt.get_cmap(colormap)(0.2)
                if load_pickle:
                    color = plt.get_cmap(colormap)(0.4)

                name = f"{root}_{spectrum.name}.png"
                pairs = spectrum.bin_pairs
                npairs = len(pairs)
                bins1 = np.transpose(pairs)[0]
                bins2 = np.transpose(pairs)[1]
                nbins1 = np.max(bins1)
                nbins2 = np.max(bins2)
                if npairs == 0:
                    continue

                # Choose different figure sizes depending on the number of redshift bins
                if not all(bins1 == bins2):
                    fig, ax = plt.subplots(nbins2, nbins1, figsize=(
                        1.6*nbins1, 1.6*nbins2), sharey=True, sharex=True)

                if all(bins1 == bins2):
                    # Autocorrelation only will have a different figure size and structure
                    fig, ax = plt.subplots(1, nbins1, figsize=(
                        1.6*nbins1, 1.4), sharey=True, sharex=True)
                    ax = np.diag(ax)

                if load_pickle:
                    # If we are continuing a plot to compare different files, load the fig and axes objects
                    name_pickle = f"{root}_{spectrum.name}.pickle"
                    fig = pl.load(open(name_pickle, 'rb'))
                    if remove_pickle:
                        os.system('rm %s' % name_pickle)
                    ax = fig.axes
                    ax = np.array(ax)
                    if len(ax) == nbins1:
                        ax = np.diag(ax)
                    else:
                        ax = np.reshape(ax, (nbins2, nbins1))

                for k, pair in enumerate(pairs):

                    i, j = pair
                    theta, xi = spectrum.get_pair(i, j)
                    error = spectrum.get_error(i, j)

                    ax[j-1][i-1].errorbar(theta, abs(xi), yerr=error, fmt=mtype, capsize=1.5,
                                          markersize=3, color=color, mec=color, elinewidth=1., label=label_legend)

                    ax[j-1][i-1].text(0.85, 0.85, f"{i},{j}", horizontalalignment='center',
                                      verticalalignment='center', transform=ax[j-1][i-1].transAxes, fontsize=12)
                    ax[j-1][i-1].set_xscale('log', nonposx='clip')
                    ax[j-1][i-1].set_yscale('log', nonposy='clip')
                    ax[j-1][i-1].xaxis.set_major_formatter(
                        ticker.FormatStrFormatter('$%d$'))
                    if blind_yaxis:
                        ax[j-1][i-1].yaxis.set_ticklabels([])

                    if (not all(bins1 == bins2)) & (j == nbins2):
                        ax[j-1][i-1].set_xlabel(r"$\theta$ [arcmin]")
                    if all(bins1 == bins2):
                        ax[j-1][i-1].set_xlabel(r"$\theta$ [arcmin]")
                    if i == 1:
                        ax[j-1][i-1].set_ylabel(label)

                    if not save_pickle:
                        if (not all(bins1 == bins2)) & (corr_type != 'gt') & (j > i):
                            fig.delaxes(ax[i-1, j-1])

                if not save_pickle:
                    plt.legend(prop={'size': 5})
                    savefig(name)
                if save_pickle:
                    name_pickle = f"{root}_{spectrum.name}.pickle"
                    pl.dump(fig, file(name_pickle, 'w'))

                plt.close()
            plt.close('all')

        if plot_kernel:
            for kernel in self.kernels:
                name = f"{root}_{kernel.name}.png"
                plt.figure()
                fig, ax = plt.subplots(1, 1, figsize=(5, 3))
                for i, nz in enumerate(kernel.nzs):
                    color = color = plt.get_cmap(colormap)(0.2*i)
                    ax.plot(kernel.z, nz, lw=1.5, color=color)
                    ax.fill_between(kernel.z, 0, nz, color=color, alpha=0.2)
                ax.set_xlabel('Redshift')
                ax.set_ylabel('Normalized counts')
                ax.set_xlim(0, 2)
                ax.set_ylim(bottom=0)
                plt.tight_layout()
                savefig(name)

        if plot_1pt:
            for obs in self.nobs:
                name = f"{root}_{obs.name}.png"
                plt.figure()
                fig, ax = plt.subplots(1, 1, figsize=(5, 3))
                for i, nob in enumerate(obs.nobs):
                    color = color = plt.get_cmap(colormap)(0.2*i)
                    ax.plot(obs.obs[i], nob, lw=1.5, color=color)
                ax.set_xlabel('Observable')
                ax.set_ylabel('Observable function')
                ax.set
                ax.set_xscale('log')
                ax.set_yscale('log')
                plt.tight_layout()
                savefig(name)

        if plot_cov:
            def corrmatrix(cov):
                cov = np.mat(cov)
                D = np.diag(np.sqrt(np.diag(cov)))
                d = np.linalg.inv(D)
                corr = d*cov*d
                return corr

            name = "{}_{}.png".format(root, 'cov')
            cov = self.covmat
            ncov1 = len(cov)
            ncov2 = len(cov[0])

            corr = corrmatrix(cov)
            if cov_vmin is None:
                cov_vmin = np.min(corr)

            figsize1 = 1.22222222227*ncov1/100.
            figsize2 = ncov2/100.
            fig, ax = plt.subplots(1, 1, figsize=(figsize1, figsize2))
            im = ax.imshow(corr, interpolation='nearest',
                           aspect='auto', origin='lower', vmin=cov_vmin, vmax=1., cmap=colormap+'_r')
            cbar = fig.colorbar(im)

            # Get labels to put in the title
            labels = ''
            for spectrum in self.spectra:
                corr_type, label, color = corr_names(spectrum)
                if corr_type is not None:
                    labels = labels + label + ' $|$ '
            labels = labels[:-4]
            ax.set_title(labels)

            # Plot lines to divide covariance
            lengths = self.covmat_info.lengths
            pos_lines = [0]
            for i in range(len(lengths)):
                pos_lines.append(pos_lines[i] + lengths[i])
            pos_lines = pos_lines[1:-1]
            for line in pos_lines:
                ax.axvline(x=line, c='k', lw=1, ls='-')
                ax.axhline(y=line, c='k', lw=1, ls='-')

            savefig(name)


class SpectrumCovarianceBuilder:
    """
    This class helps you ensure that the ordering between a set of data points and
    their covariance is consistently maintained.  You add data points to it one by one.,
    in the order that they appear in the covariance.

    Here is an example using ther CFHTLenS revisited files:

    >>> theta_values = [1.41, 2.79, 5.53, 11.0, 21.7, 43.0, 85.2]
    >>> types = {
        'xip': twopoint.Types.galaxy_shear_plus_real,
        'xim': twopoint.Types.galaxy_shear_minus_real,
    }
    >>> kernel='NZ_SOURCE'
    >>> builder = twopoint.SpectrumCovarianceBuilder()
    >>> i_bin, data, jk_err, sim_err = np.loadtxt("./xipm_cfhtlens_regcomb_blind1_passfields_athenazsj.dat").T
    >>> i = 0
    >>> for bin1 in range(7):
        for bin2 in range(bin1,7):
            for name in ['xip', 'xim']:
                stype = types[name]
                for angbin in range(1,8):
                    ang = theta_values[angbin-1]
                    value = data[i]
                    i+=1
                    builder.add_data_point(kernel,kernel,stype,stype,bin1+1,bin2+1,ang,angbin,value)
    >>> names = {builder.types[0]:"xip", builder.types[1]:"xim"}
    >>> builder.set_names(names)
    >>> spectra, covmat_info = builder.generate(covmat,"arcmin")
    """

    def __init__(self):
        self.kernel1 = []
        self.kernel2 = []
        self.bin1 = []
        self.bin2 = []
        self.type1 = []
        self.type2 = []
        self.ang = []
        self.angbin = []
        self.value = []
        self.names = None
        self.types = []
        self.total_length = 0
        self.total_length_1pt = 0
        self.nobs = []
        self.obs = []
        self.obs_name = []
        self.nbin = []

    def add_data_point(self, kernel1, kernel2, type1, type2, bin1, bin2, ang, angbin, value):
        self.kernel1.append(kernel1)
        self.kernel2.append(kernel2)
        self.type1.append(type1)
        self.type2.append(type2)
        self.bin1.append(bin1)
        self.bin2.append(bin2)
        self.ang.append(ang)
        self.angbin.append(angbin)
        self.value.append(value)
        self.total_length += 1

        spec = kernel1, kernel2, type1, type2
        if spec not in self.types:
            self.types.append(spec)

    def add_one_point(self, obs_name, nbin, x, angbin, values):
        self.obs_name.append(obs_name)
        self.nbin.append(nbin)
        self.obs.append(x)
        self.nobs.append(values)
        #self.angbin.append(angbin)
        self.total_length_1pt += 1

    def _freeze(self):
        self.bin1 = np.array(self.bin1)
        self.bin2 = np.array(self.bin2)
        self.ang = np.array(self.ang)
        self.angbin = np.array(self.angbin)
        self.value = np.array(self.value)
        self.nbin = np.array(self.nbin)
        self.obs = np.array(self.obs)
        self.nobs = np.array(self.nobs)

    def set_names(self, names):
        for t in self.types:
            if t not in names:
                raise ValueError(
                    f"Please provide name for spectrum of type: {t}")
        self.names = names

    def generate(self, covariance, angle_unit):
        if self.names is None:
            raise ValueError(
                "Please provide names for each type in self.types first")
        self._freeze()
        assert covariance.shape == (self.total_length + self.total_length_1pt, self.total_length + self.total_length_1pt)
        # get all the unique pairs of type1,type2.  maintain order
        # shouldn't be more than a few types so the list membership test is fast
        master_index_vector = []
        spectra = []
        onepoint = []
        for kernel1, kernel2, type1, type2 in self.types:
            kernels = (kernel1, kernel2)
            types = (type1, type2)
            spectrum_index_vector = []
            for i in range(self.total_length):
                k1 = self.kernel1[i]
                k2 = self.kernel2[i]
                t1 = self.type1[i]
                t2 = self.type2[i]
                if (type1 != t1) or (type2 != t2) or (kernel1 != k1) or (kernel2 != k2):
                    continue
                master_index_vector.append(i)
                spectrum_index_vector.append(i)
            bins = (self.bin1[spectrum_index_vector],
                    self.bin2[spectrum_index_vector])
            angular_bin = self.angbin[spectrum_index_vector]
            value = self.value[spectrum_index_vector]
            angle = self.ang[spectrum_index_vector]
            name = self.names[(kernel1, kernel2, type1, type2)]
            windows = "SAMPLE"
            spectrum = SpectrumMeasurement(
                name, bins, types, kernels, windows, angular_bin, value,
                angle=angle, angle_unit=angle_unit)
            spectra.append(spectrum)

        for name in np.unique(self.obs_name):
            spectrum_index_vector = []
            for i in range(self.total_length,self.total_length_1pt+self.total_length):
                master_index_vector.append(i)
            for i in range(self.total_length_1pt):
                spectrum_index_vector.append(i)
            bins = self.nbin[spectrum_index_vector]
            angular_bin = self.obs[spectrum_index_vector]
            value = self.nobs[spectrum_index_vector]
            onepoint.append(OnePointMeasurement('1pt', [angular_bin], [value]))

        reordered_covariance = covariance[master_index_vector][:, master_index_vector]
        covmat_info = CovarianceMatrixInfo("COVMAT", [s.name for s in spectra] + [o.name for o in onepoint],
                                           [len(s) for s in spectra] + [o.nsample for o in onepoint], reordered_covariance)

        return spectra, covmat_info


if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    output_root = sys.argv[2]
    T = TwoPointFile.from_fits(filename, covmat_name=None)
    T.plots(output_root)
